from octis.evaluation_metrics.diversity_metrics import WordEmbeddingsInvertedRBO, \
    WordEmbeddingsInvertedRBOCentroid, InvertedRBO
import numpy as np
from itertools import combinations
from scipy.spatial.distance import cosine
from octis.evaluation_metrics.metrics import AbstractMetric
from gensim.models import KeyedVectors
import gensim.downloader as api


class WordEmbeddingsRBOMatch(WordEmbeddingsInvertedRBO):
    def __init__(self, word2vec_path=None, binary=True, normalize=True, weight=0.9, topk=10):
        """
        Initialize metric WERBO-Match

        Parameters
        ----------
        :param topk: top k words on which the topic diversity will be computed
        :param word2vec_path: word embedding space in gensim word2vec format
        :param weight: Weight of each agreement at depth d. When set to 1.0, there is no weight, the rbo returns to
        average overlap. (Default 0.9)
        :param normalize: if true, normalize the cosine similarity
        """
        super().__init__(word2vec_path=word2vec_path, binary=binary, normalize=normalize, weight=weight, topk=topk)

    def score(self, model_output):
        """
        Retrieves the score of the metric

        :return WERBO-M
        """
        return 1 - super(WordEmbeddingsRBOMatch, self).score(model_output)


class WordEmbeddingsRBOCentroid(WordEmbeddingsInvertedRBOCentroid):
    def __init__(self, word2vec_path=None, binary=True, normalize=True, weight=0.9, topk=10):
        """
        Initialize metric WERBO-Centroid

        Parameters
        ----------
        :param topk: top k words on which the topic diversity will be computed
        :param word2vec_path: word embedding space in gensim word2vec format
        :param weight: Weight of each agreement at depth d. When set to 1.0, there is no weight, the rbo returns to
        average overlap. (Default 0.9)
        :param normalize: if true, normalize the cosine similarity
        """
        super().__init__(word2vec_path=word2vec_path, binary=binary, normalize=normalize, weight=weight, topk=topk)

    def score(self, model_output):
        """
        Retrieves the score of the metric

        :return WERBO-C
        """
        return 1 - super(WordEmbeddingsRBOCentroid, self).score(model_output)


class WordEmbeddingsPairwiseSimilarity(AbstractMetric):
    def __init__(self, word2vec_path=None, topk=10):
        """
        Initialize metric WE pairwise similarity

        Parameters
        ----------
        :param topk: top k words on which the topic diversity will be computed
        :param word2vec_path: word embedding space in gensim word2vec format
        """
        super().__init__()
        if word2vec_path is None:
            self.wv = api.load('word2vec-google-news-300')
        else:
            self.wv = KeyedVectors.load_word2vec_format( word2vec_path, binary=self.binary)

        self.topk = topk

    def score(self, model_output):
        """
        Retrieves the score of the metric

        :return WEPS
        """
        topics = model_output['topics']
        if self.topk > len(topics[0]):
            raise Exception('Words in topics are less than topk')
        else:
            count = 0
            sum_sim = 0
            for list1, list2 in combinations(topics, 2):
                word_counts = 0
                sim = 0
                for word1 in list1[:self.topk]:
                    for word2 in list2[:self.topk]:
                        if word1 in self.wv.wv.vocab and word2 in self.wv.wv.vocab:
                            sim = sim + self.wv.similarity(word1, word2)
                            word_counts = word_counts + 1
                sim = sim / word_counts
                sum_sim = sum_sim + sim
                count = count + 1
            return sum_sim / count


class WordEmbeddingsCentroidSimilarity(AbstractMetric):
    def __init__(self, word2vec_path=None, topk=10):
        """
        Initialize metric WE centroid similarity

        Parameters
        ----------
        :param topk: top k words on which the topic diversity will be computed
        :param word2vec_path: word embedding space in gensim word2vec format
        """
        super().__init__()
        if word2vec_path is None:
            self.wv = api.load('word2vec-google-news-300')
        else:
            self.wv = KeyedVectors.load_word2vec_format(word2vec_path, binary=self.binary)
        self.topk = topk

    def score(self, model_output):
        """
        Retrieves the score of the metric

        :return WECS
        """
        topics = model_output['topics']
        if self.topk > len(topics[0]):
            raise Exception('Words in topics are less than topk')
        else:
            sim = 0
            count = 0
            for list1, list2 in combinations(topics, 2):
                centroid1 = np.zeros(self.wv.vector_size)
                centroid2 = np.zeros(self.wv.vector_size)
                count1, count2 = 0, 0
                for word1 in list1[:self.topk]:
                    if word1 in self.wv.wv.vocab:
                        centroid1 = centroid1 + self.wv[word1]
                        count1 += 1
                for word2 in list2[:self.topk]:
                    if word2 in self.wv.wv.vocab:
                        centroid2 = centroid2 + self.wv[word2]
                        count2 += 1
                centroid1 = centroid1 / count1
                centroid2 = centroid2 / count2
                sim = sim + (1 - cosine(centroid1, centroid2))
                count += 1
            return sim / count


def get_word2index(list1, list2):
    words = set(list1)
    words = words.union(set(list2))
    word2index = {w: i for i, w in enumerate(words)}
    return word2index


class WordEmbeddingsWeightedSumSimilarity(AbstractMetric):
    def __init__(self, id2word, word2vec_path=None, topk=10):
        """
        Initialize metric WE Weighted Sum similarity

        :param id2word: dictionary mapping each id to the word of the vocabulary
        :param topk: top k words on which the topic diversity will be computed
        :param word2vec_path: word embedding space in gensim word2vec format
        """
        super().__init__()
        if word2vec_path is None:
            self.wv = api.load('word2vec-google-news-300')
        else:
            self.wv = KeyedVectors.load_word2vec_format(word2vec_path, binary=self.binary)
        self.topk = topk

    def score(self, model_output):
        """
        Retrieves the score of the metric

        :return WESS
        """
        beta = model_output['topic-word-distribution']

        wess = 0
        count = 0
        for i, j in combinations(range(len(beta)), 2):
            centroid1 = np.zeros(self.wv.vector_size)
            weights = 0
            for id_beta, w in enumerate(beta[i]):
                centroid1 = centroid1 + self.wv[self.id2word[id_beta]] * w
                weights += w
            centroid1 = centroid1 / weights
            centroid2 = np.zeros(self.wv.vector_size)
            weights = 0
            for id_beta, w in enumerate(beta[i]):
                centroid2 = centroid2 + self.wv[self.id2word[id_beta]] * w
                weights += w
            centroid2 = centroid2 / weights
            wess += cosine(centroid1, centroid2)
        return wess / count


class RBO(InvertedRBO):
    def __init__(self, weight=0.9, topk=10):
        """
        Initialize metric Ranked-biased Overlap

        Parameters
        ----------
        :param topk: top k words on which the topic diversity will be computed
        :param weight: Weight of each agreement at depth d. When set to 1.0, there is no weight, the rbo returns to
        average overlap. (Default 0.9)
        """
        super().__init__(weight=weight, topk=topk)

    def score(self, model_output):
        """
        Retrieves the score of the metric

        :return RBO
        """
        return 1 - super(RBO, self).score(model_output)


class PairwiseJaccardSimilarity(AbstractMetric):
    def __init__(self, topk=10):
        """
        Initialize metric Pairwise Jaccard Similarity

        Parameters
        ----------
        :param topk: top k words on which the topic diversity will be computed

        """
        super().__init__()
        self.topk = topk

    def score(self, model_output):
        """
        Retrieves the score of the metric

        :return PJS
        """
        topics = model_output['topics']
        sim = 0
        count = 0
        for list1, list2 in combinations(topics, 2):
            intersection = len(list(set(list1[:self.topk]).intersection(list2[:self.topk])))
            union = (len(list1[:self.topk]) + len(list2[:self.topk])) - intersection
            count = count + 1
            sim = sim + (float(intersection) / union)
        return sim / count




#!/usr/bin/env python

"""Tests for `octis` package."""

import pytest

from click.testing import CliRunner
from octis.evaluation_metrics.topic_significance_metrics import *
from octis.evaluation_metrics.classification_metrics import F1Score, PrecisionScore
from octis.evaluation_metrics.classification_metrics import AccuracyScore, RecallScore
from octis.evaluation_metrics.diversity_metrics import TopicDiversity, InvertedRBO, KLDivergence, LogOddsRatio, \
    WordEmbeddingsInvertedRBO
from octis.evaluation_metrics.similarity_metrics import WordEmbeddingsRBOMatch, PairwiseJaccardSimilarity, RBO, \
    WordEmbeddingsCentroidSimilarity

from octis.evaluation_metrics.coherence_metrics import *
from octis.dataset.dataset import Dataset
from octis.models.LDA import LDA

import os


@pytest.fixture
def root_dir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def dataset(root_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(root_dir + "/../preprocessed_datasets/" + '/M10')
    return dataset


@pytest.fixture
def model_output(dataset):
    model = LDA(num_topics=3, iterations=5)
    output = model.train_model(dataset)
    return output


def test_f1score(dataset, model_output):
    metric = F1Score(dataset=dataset)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float


def test_accuracyscore(dataset, model_output):
    metric = AccuracyScore(dataset=dataset)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float


def test_precisionscore(dataset, model_output):
    metric = PrecisionScore(dataset=dataset)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float


def test_recallscore(dataset, model_output):
    metric = RecallScore(dataset=dataset)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float


def test_svm_persistency(dataset, model_output):
    metric = F1Score(dataset=dataset)
    metric.score(model_output)
    metric = AccuracyScore(dataset=dataset)
    metric.score(model_output)
    assert metric.same_svm
    metric = F1Score(dataset=dataset, average="macro")
    metric.score(model_output)
    assert not metric.same_svm


def test_npmi_coherence_measures(dataset, model_output):
    metric = Coherence(topk=10, texts=dataset.get_corpus())
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert -1 <= score <= 1

def test_we_coherence_measures(dataset, model_output):
    metric = WECoherenceCentroid(topk=5)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == np.float32 or type(score) == float
    assert -1 <= score <= 1

    metric = WECoherencePairwise(topk=10)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == np.float32 or type(score) == float
    assert -1 <= score <= 1


def test_diversity_measures(dataset, model_output):
    metric = TopicDiversity(topk=10)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert 0 <= score <= 1

    metric = KLDivergence()
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert 0 <= score <= 1

    metric = LogOddsRatio()
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert 0 <= score <= 1

    metric = WordEmbeddingsInvertedRBO(normalize=True)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert 0 <= score <= 1


def test_similarity_measures(dataset, model_output):
    metric = RBO(topk=10)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert 0 <= score <= 1

    metric = WordEmbeddingsRBOMatch(topk=10, normalize=True)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert 0 <= score <= 1

    metric = PairwiseJaccardSimilarity(topk=10)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert 0 <= score <= 1

    metric = WordEmbeddingsCentroidSimilarity(topk=10)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert 0 <= score <= 1


def test_irbo(dataset, model_output):
    metric = InvertedRBO(topk=10)
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert 0 <= score <= 1


def test_kl_b(dataset, model_output):
    metric = KL_background()
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert score >= 0


def test_kl_v(dataset, model_output):
    metric = KL_vacuous()
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert score >= 0


def test_kl_u(dataset, model_output):
    metric = KL_uniform()
    score = metric.score(model_output)
    assert type(score) == np.float64 or type(score) == float
    assert score >= 0



# Utils
import json
import time
from pathlib import Path

# utils from skopt and sklearn
from sklearn.gaussian_process.kernels import *
from skopt.space.space import *

from octis.dataset.dataset import Dataset
# utils from other files of the framework
from octis.models.model import save_model_output
from octis.optimization.optimizer_evaluation import OptimizerEvaluation
from octis.optimization.optimizer_tool import choose_optimizer, early_condition, load_model, select_metric
from octis.optimization.optimizer_tool import load_search_space, plot_bayesian_optimization, plot_model_runs


class Optimizer:
    """
    Class Optimizer to perform Bayesian Optimization on Topic Model
    """

    def optimize(self, model, dataset, metric, search_space, extra_metrics=None,
                 number_of_call=5, n_random_starts=1,
                 initial_point_generator="lhs",  # work only for version skopt 8.0!!!
                 optimization_type='Maximize', model_runs=5, surrogate_model="RF",
                 kernel=1.0 * Matern(length_scale=1.0, length_scale_bounds=(1e-1, 10.0), nu=1.5),
                 acq_func="LCB", random_state=False, x0=None, y0=None,
                 save_models=True, save_step=1, save_name="result", save_path="results/", early_stop=False,
                 early_step=5,
                 plot_best_seen=False, plot_model=False, plot_name="B0_plot", log_scale_plot=False, topk=10):
        """
        Perform hyper-parameter optimization for a Topic Model

        :param model: model with hyperparameters to optimize
        :type model: OCTIS Topic Model
        :param dataset: dataset for the model dataset
        :type dataset: OCTIS dataset
        :param metric: metric used for the optimization
        :type metric: OCTIS metric
        :param search_space: a dictionary of hyperparameters to optimize (each parameter is defined as a skopt space)
        :type search_space: skopt space object
        :param extra_metrics: list of extra-metrics to compute during the optimization
        :type extra_metrics: list of metrics, optional
        :param number_of_call: number of evaluations of metric
        :type number_of_call: int, optional
        :param n_random_starts: number of evaluations of metric with random points before approximating it with surrogate model
        :type n_random_starts: int, optional
        :param initial_point_generator: set an initial point generator. Can be either "random", "sobol", "halton" ,"hammersly","lhs"
        :type initial_point_generator: str, optional
        :param optimization_type: Set "Maximize" if you want to maximize metric, "Minimize" if you want to minimize
        :type optimization type: str, optional
        :param model_runs:
        :type: int, optional
        :param surrogate_model: set a surrogate model. Can be either "GP" (Gaussian Process), "RF" (Random Forest) or "ET" (Extra-Tree)
        :type: str, optional
        :param kernel: set a kernel function
        :param acq_func: Function to minimize over the surrogate model. Can be either: "LCB" (Lower Confidence Bound), "EI" (Expected improvement) OR "PI" (Probability of Improvement)
        :type: str, optional
        :param random_state: Set random state to something other than None for reproducible results.
        :type: int, optional
        :param x0: List of initial input points.
        :type: list, optional
        :param y0: Evaluation of initial input points.
        :type: list, optional
        :param save_models: if 'True' save all the topic models generated during the optimization process
        :type: bool, optional
        :param save_step: decide how much to save the results of the optimization
        :type: int, optional
        :param save_name: name of the file where the results of the optimization will be saved
        :type: str, optional
        :param save_path: Path where the results of the optimization (json file) will be saved
        :type save_path: str, optional
        :param early_stop: if "True" stop the optimization if there is no improvement after early_step evaluations
        :type early_stop: bool, optional
        :param early_step: number of iterations with no improvement after which optimization will be stopped (if early_stop is True)
        :type early_step: int, optional
        :param plot_best_seen: If "True" save a convergence plot of the result of a Bayesian_optimization (i.e. the best seen for each iteration)
        :type plot_best_seen: bool, optional
        :param plot_model: If "True" save the boxplot of all the model runs
        :type plot_model: bool, optional
        :param plot_name: Set the name of the plots (best_seen and model_runs).
        :type plot_name: str, optional
        :param log_scale_plot: if "True" use the logarithmic scale for the plots.
        :type log_scale_plot: bool, optional
        :param topk:
        :type topk: int, optional
        :return: OptimizerEvaluation object
        :rtype: class
        """
        # Set the attributes
        if extra_metrics is None:
            extra_metrics = []
        if y0 is None:
            y0 = []
        if x0 is None:
            x0 = dict()
        self.model = model
        self.dataset = dataset
        self.metric = metric
        self.search_space = search_space
        self.extra_metrics = extra_metrics
        self.optimization_type = optimization_type
        self.number_of_call = number_of_call
        self.n_random_starts = n_random_starts
        self.initial_point_generator = initial_point_generator
        self.model_runs = model_runs
        self.surrogate_model = surrogate_model
        self.kernel = kernel
        self.acq_func = acq_func
        self.random_state = random_state
        self.x0 = x0
        self.y0 = y0
        self.save_path = save_path
        self.save_step = save_step
        self.save_name = save_name
        self.save_models = save_models
        self.early_stop = early_stop
        self.early_step = early_step
        self.plot_model = plot_model
        self.plot_best_seen = plot_best_seen
        self.plot_name = plot_name
        self.log_scale_plot = log_scale_plot
        self.topk = topk

        self.hyperparameters = list(sorted(self.search_space.keys()))
        self.dict_model_runs = dict()
        self.number_of_previous_calls = 0
        self.current_call = 0
        self.time_eval = []

        self.name_optimized_metric = metric.__class__.__name__
        self.dict_model_runs[self.name_optimized_metric] = dict()

        # Info about extra metrics
        i = 0
        self.extra_metric_names = []
        for extra_metric in extra_metrics:
            self.extra_metric_names.append(str(i) + '_' + extra_metric.__class__.__name__)
            self.dict_model_runs[self.extra_metric_names[i]] = dict()
            i = i + 1

        # Control about the correctness of BO parameters
        if self._check_bo_parameters() == -1:
            print("ERROR: wrong initialitation of BO parameters")
            return None

        # Create the directory where the results are saved
        Path(self.save_path).mkdir(parents=True, exist_ok=True)

        # Initialize the directories about model_runs
        if self.save_models:
            self.model_path_models = self.save_path + "models/"
            Path(self.model_path_models).mkdir(parents=True, exist_ok=True)

        # Choice of the optimizer
        opt = choose_optimizer(self)

        # Perform Bayesian Optimization
        results = self._optimization_loop(opt)

        return results

    def resume_optimization(self, name_path, extra_evaluations=0):
        """
        Restart the optimization from the json file.

        :param name_path: path of the json file
        :type name_path: str
        :param extra_evaluations: extra iterations for the BO optimization
        :type extra_evaluations: int
        :return: object with the results of the optimization
        :rtype: object
        """

        # Restore of the parameters
        res, opt = self._restore_parameters(name_path)

        # Set the number of total calls
        self.number_of_call = self.number_of_call + extra_evaluations

        # Check if there are other iterations to do
        if self.number_of_previous_calls == self.number_of_call:
            return OptimizerEvaluation(self, BO_results=res)

        # Control about the correctness of BO parameters
        if self._check_bo_parameters() == -1:
            print("ERROR: wrong initialization of BO parameters")
            return None

        results = self._optimization_loop(opt)

        return results

    def _objective_function(self, hyperparameter_values):
        """
        Evaluate the objective function

        :param hyperparameter_values: hyper-parameters of the Topic Model
        :type hyperparameter_values: list
        :return: value of the objective function
        :rtype: float
        """
        # Retrieve parameters labels
        params = {}
        for i in range(len(self.hyperparameters)):
            params[self.hyperparameters[i]] = hyperparameter_values[i]

        # Compute the score of the hyper-parameter configuration
        different_model_runs = []
        different_model_runs_extra_metrics = [[] for i in range(len(self.extra_metrics))]

        for i in range(self.model_runs):

            # Prepare model
            model_output = self.model.train_model(self.dataset, params,
                                                  self.topk)
            # Score of the model
            score = self.metric.score(model_output)
            different_model_runs.append(score)

            # Update of the extra metric values
            for j, extra_metric in enumerate(self.extra_metrics):
                different_model_runs_extra_metrics[j].append(extra_metric.score(model_output))

            # Save the model for each run
            if self.save_models:
                name = str(self.current_call) + "_" + str(i)
                save_model_path = self.model_path_models + name
                save_model_output(model_output, save_model_path)

        # Update of the dictionaries
        self.dict_model_runs[self.name_optimized_metric][
            'iteration_' + str(self.current_call)] = different_model_runs

        for j, extra_metric in enumerate(self.extra_metrics):
            self.dict_model_runs[self.extra_metric_names[j]]['iteration_' + str(self.current_call)] = \
                different_model_runs_extra_metrics[j]

        # The output for BO is the median over different_model_runs
        result = np.median(different_model_runs)

        if self.optimization_type == 'Maximize':
            result = - result

        # Boxplot for matrix_model_runs
        if self.plot_model:
            name_plot = self.save_path + self.plot_name + "_model_runs_" + self.name_optimized_metric
            plot_model_runs(self.dict_model_runs[self.name_optimized_metric], self.current_call, name_plot)

            # Boxplot of extrametrics (if any)
            for j in range(len(self.extra_metrics)):
                name_plot = self.save_path + self.plot_name + "_model_runs_" + self.extra_metric_names[j]
                plot_model_runs(self.dict_model_runs[self.extra_metric_names[j]], self.current_call, name_plot)

        return result

    def _optimization_loop(self, opt):
        """
        Perform the optimization through Bayesian Optimization

        :return: result of the optimization
        :rtype: class
        """
        results = None
        # For loop to perform Bayesian Optimization
        for i in range(self.number_of_previous_calls, self.number_of_call):

            # Next point proposed by BO and evaluation of the objective function
            print("Current call: ", self.current_call)
            start_time = time.time()

            # Next point proposed by BO and evaluation of the objective function
            if i < self.lenx0:
                next_x = [self.x0[name][i] for name in self.hyperparameters]
                # next_x = self.x0[i]
                if len(self.y0) == 0:
                    f_val = self._objective_function(next_x)
                else:
                    self.dict_model_runs[self.name_optimized_metric]['iteration_' + str(i)] = self.y0[i]
                    f_val = -self.y0[i] if self.optimization_type == 'Maximize' else self.y0[i]

            else:
                next_x = opt.ask()
                f_val = self._objective_function(next_x)

            # Update the opt using (next_x,f_val)
            res = opt.tell(next_x, f_val)

            # Update the computational time for next_x (BO+Function evaluation)
            end_time = time.time()
            total_time_function = end_time - start_time
            self.time_eval.append(total_time_function)

            # Plot best seen
            if self.plot_best_seen:
                plot_bayesian_optimization(res.func_vals, self.save_path + self.plot_name + "_best_seen",
                                           self.log_scale_plot, conv_max=self.optimization_type == 'Maximize')

            # Create an object related to the BO optimization
            results = OptimizerEvaluation(self, BO_results=res)

            # Save the object
            if i % self.save_step == 0:
                name_json = self.save_path + self.save_name + ".json"
                results.save(name_json)

            # Early stop condition
            if i >= len(self.x0) and self.early_stop and early_condition(res.func_vals, self.early_step,
                                                                         self.n_random_starts):
                print("Stop because of early stopping condition")
                break

            # Update current_call
            self.current_call = self.current_call + 1

        return results

    def _load_metric(self, optimization_object, dataset):
        """
        Load the metric from the json file, useful for the resume method

        :param optimization_object: dictionary of the information saved during the optimization
        :type optimization_object: dict
        :param dataset: the considered dataset
        :type dataset: OCTIS dataset object
        """
        # Optimized Metric
        self.name_optimized_metric = optimization_object['metric_name']
        metric_parameters = optimization_object['metric_attributes']

        if self.name_optimized_metric.startswith('Coherence'):
            metric_parameters.update({'texts': dataset.get_corpus()})
        if self.name_optimized_metric.startswith('F1Score'):
            metric_parameters.update({'dataset': dataset})

        self.metric = select_metric(metric_parameters, self.name_optimized_metric)

        # Extra metrics
        self.extra_metrics = []
        self.extra_metric_names = optimization_object['extra_metric_names']
        dict_extra_metric_parameters = optimization_object['extra_metric_attributes']

        for name in self.extra_metric_names:
            metric_parameters = dict_extra_metric_parameters[name]
            if 'Coherence' in name:
                metric_parameters.update({'texts': dataset.get_corpus()})
            if 'F1Score' in name:
                metric_parameters.update({'dataset': dataset})

            metric = select_metric(metric_parameters, name[2:])
            self.extra_metrics.append(metric)

    def _restore_parameters(self, name_path):
        """
        Restore the BO parameters  from the json file

        :param name_path: name of the json file
        :type name_path: str
        :return: result of BO optimization (scikit-optimize object), surrogate model (scikit-learn object)
        :rtype: tuple
        """

        # Load the previous results
        with open(name_path, 'rb') as file:
            optimization_object = json.load(file)

        self.search_space = load_search_space(optimization_object["search_space"])
        self.acq_func = optimization_object["acq_func"]
        self.surrogate_model = optimization_object["surrogate_model"]
        self.kernel = eval(optimization_object["kernel"])
        self.optimization_type = optimization_object["optimization_type"]
        self.model_runs = optimization_object["model_runs"]
        self.save_models = optimization_object["save_models"]
        self.save_step = optimization_object["save_step"]
        self.save_name = optimization_object["save_name"]
        self.save_models = optimization_object["save_models"]
        self.save_path = optimization_object["save_path"]
        self.early_stop = optimization_object["early_stop"]
        self.early_step = optimization_object["early_step"]
        self.plot_model = optimization_object["plot_model"]
        self.plot_best_seen = optimization_object["plot_best_seen"]
        self.plot_name = optimization_object["plot_name"]
        self.log_scale_plot = optimization_object["log_scale_plot"]
        self.random_state = optimization_object["random_state"]
        self.dict_model_runs = optimization_object['dict_model_runs']
        self.number_of_previous_calls = optimization_object['current_call'] + 1
        self.current_call = optimization_object['current_call'] + 1
        self.number_of_call = optimization_object['number_of_call']
        self.save_path = optimization_object['save_path']
        self.x0 = optimization_object['x0']
        self.y0 = optimization_object['y0']
        self.n_random_starts = optimization_object['n_random_starts']
        self.initial_point_generator = optimization_object['initial_point_generator']
        self.topk = optimization_object['topk']
        self.time_eval = optimization_object["time_eval"]
        res = None

        # Load the dataset
        dataset = Dataset()
        if not optimization_object["is_cached"]:
            dataset.load_custom_dataset_from_folder(optimization_object["dataset_path"])
        else:
            dp = optimization_object["dataset_path"][:-(len(optimization_object["dataset_name"]) + len("_py3.pkz"))]
            dataset.fetch_dataset(optimization_object["dataset_name"], data_home=dp)

        self.dataset = dataset

        # Load the metric
        self._load_metric(optimization_object, dataset)

        # Load the model
        self.model = load_model(optimization_object)
        # Creation of the hyperparameters
        self.hyperparameters = list(sorted(self.search_space.keys()))
        # Choice of the optimizer
        opt = choose_optimizer(self)

        # Update number_of_call for restarting
        for i in range(self.number_of_previous_calls):
            next_x = [optimization_object["x_iters"][key][i] for key in self.hyperparameters]
            f_val = -optimization_object["f_val"][i] if self.optimization_type == 'Maximize' else \
                optimization_object["f_val"][i]
            res = opt.tell(next_x, f_val)

            # Create the directory where the results are saved
        Path(self.save_path).mkdir(parents=True, exist_ok=True)

        self.model_path_models = self.save_path + "models/"

        return res, opt

    def _check_bo_parameters(self):
        """
        Check the correctness of BO parameters

        :return: -1 if there is an error, 0 otherwise
        :rtype: bool
        """

        if self.optimization_type not in ['Maximize', 'Minimize']:
            print("Error: optimization type must be Maximize or Minimize")
            return -1

        if self.surrogate_model not in ['RF', 'RS', 'GP', 'ET']:
            print("Error: surrogate model must be RF, ET, RS or GP")
            return -1

        if self.acq_func not in ['PI', 'EI', 'LCB']:
            print("Error: acquisition function must be PI, EI or LCB")
            return -1

        if self.number_of_call <= 0:
            print("Error: number_of_call can't be <= 0")
            return -1

        if self.number_of_call - len(self.x0) <= 0:
            print("Error: number_of_call is less then len(x0)")
            return None

        if not isinstance(self.model_runs, int):
            print("Error: model_run must be an integer")
            return -1

        if not isinstance(self.number_of_call, int):
            print("Error: number_of_call must be an integer")
            return -1

        if not isinstance(self.n_random_starts, int):
            print("Error: n_random_starts must be an integer")
            return -1

        if not isinstance(self.save_step, int):
            print("Error: save_step must be an integer")
            return -1

        if not isinstance(self.save_step, int):
            print("Error: save_step must be an integer")
            return -1

        if self.n_random_starts <= 0:
            print("Error: the number of initial_points must be >=1 !!!")
            return -1

        if self.initial_point_generator not in ['lhs', 'sobol', 'halton', 'hammersly', 'grid', 'random']:
            print("Error: wrong initial_point_generator")
            return -1

        if not isinstance(self.x0, dict):
            print("Error: x0 must be a dictionary!")
            return -1

        if not isinstance(self.y0, list):
            print("Error: y0 must be a dictionary!")
            return -1

        if len(self.x0) > 0:
            self.lenx0 = len(list(self.x0.values())[0])
            for i in range(len(self.x0.values())):
                lenC = len(list(self.x0.values())[i])
                if lenC != self.lenx0:
                    print("Error: dimension of x0 is not consistent!")
                    return -1

            if len(self.y0) > 0:
                if self.lenx0 != len(self.y0):
                    print("Error: different dimension for x0 and y0!")
                    return -1

        else:
            self.lenx0 = 0
            self.leny0 = 0

        if self.plot_name.endswith(".png"):
            self.plot_name = self.plot_name[:-4]

        if self.save_name.endswith(".json"):
            self.save_name = self.save_name[:-5]

        if self.save_path[-1] != '/':
            self.save_path = self.save_path + '/'

        return 0



import torch
import torch.nn.functional as F
import numpy as np
import math

from torch import nn


class ETM(nn.Module):
    def __init__(self, num_topics, vocab_size, t_hidden_size, rho_size, emb_size,
                 theta_act, embeddings=None, train_embeddings=True, enc_drop=0.5):
        super(ETM, self).__init__()
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        ## define hyperparameters
        self.num_topics = num_topics
        self.vocab_size = vocab_size
        self.t_hidden_size = t_hidden_size
        self.rho_size = rho_size
        self.enc_drop = enc_drop
        self.emb_size = emb_size
        self.t_drop = nn.Dropout(enc_drop)

        self.theta_act = self.get_activation(theta_act)

        ## define the word embedding matrix \rho
        if train_embeddings:
            self.rho = nn.Linear(rho_size, vocab_size, bias=False)
        else:
            num_embeddings, emb_size = embeddings.size()
            rho = nn.Embedding(num_embeddings, emb_size)
            self.rho = embeddings.clone().float().to(self.device)

        ## define the matrix containing the topic embeddings
        self.alphas = nn.Linear(rho_size, num_topics, bias=False)#nn.Parameter(torch.randn(rho_size, num_topics))

        ## define variational distribution for \theta_{1:D} via amortizartion
        self.q_theta = nn.Sequential(
                nn.Linear(vocab_size, t_hidden_size),  self.theta_act,
                nn.Linear(t_hidden_size, t_hidden_size), self.theta_act,
            )
        self.mu_q_theta = nn.Linear(t_hidden_size, num_topics, bias=True)
        self.logsigma_q_theta = nn.Linear(t_hidden_size, num_topics, bias=True)

    def get_activation(self, act):
        if act == 'tanh':
            act = nn.Tanh()
        elif act == 'relu':
            act = nn.ReLU()
        elif act == 'softplus':
            act = nn.Softplus()
        elif act == 'sigmoid':
            act = nn.Sigmoid()
        elif act == 'rrelu':
            act = nn.RReLU()
        elif act == 'leakyrelu':
            act = nn.LeakyReLU()
        elif act == 'elu':
            act = nn.ELU()
        elif act == 'selu':
            act = nn.SELU()
        elif act == 'glu':
            act = nn.GLU() #error using glu
        else:
            print('Defaulting to tanh activations...')
            act = nn.Tanh()
        return act

    def reparameterize(self, mu, logvar):
        """Returns a sample from a Gaussian distribution via reparameterization.
        """
        if self.training:
            std = torch.exp(0.5 * logvar)
            eps = torch.randn_like(std)
            return eps.mul_(std).add_(mu)
        else:
            return mu

    def encode(self, bows):
        """Returns paramters of the variational distribution for \theta.

        input: bows
                batch of bag-of-words...tensor of shape bsz x V
        output: mu_theta, log_sigma_theta
        """
        q_theta = self.q_theta(bows)
        if self.enc_drop > 0:
            q_theta = self.t_drop(q_theta)
        mu_theta = self.mu_q_theta(q_theta)
        logsigma_theta = self.logsigma_q_theta(q_theta)
        kl_theta = -0.5 * torch.sum(1 + logsigma_theta - mu_theta.pow(2) - logsigma_theta.exp(), dim=-1).mean()
        return mu_theta, logsigma_theta, kl_theta

    def get_beta(self):
        try:
            logit = self.alphas(self.rho.weight) # torch.mm(self.rho, self.alphas)
        except:
            logit = self.alphas(self.rho)
        beta = F.softmax(logit, dim=0).transpose(1, 0) ## softmax over vocab dimension
        return beta

    def get_theta(self, normalized_bows):
        mu_theta, logsigma_theta, kld_theta = self.encode(normalized_bows)
        z = self.reparameterize(mu_theta, logsigma_theta)
        theta = F.softmax(z, dim=-1)
        return theta, kld_theta

    def decode(self, theta, beta):
        res = torch.mm(theta, beta)
        preds = torch.log(res+1e-6)
        return preds

    def forward(self, bows, normalized_bows, theta=None, aggregate=True):
        ## get \theta
        if theta is None:
            theta, kld_theta = self.get_theta(normalized_bows)
        else:
            kld_theta = None

        ## get \beta
        beta = self.get_beta()

        ## get prediction loss
        preds = self.decode(theta, beta)
        recon_loss = -(preds * bows).sum(1)
        if aggregate:
            recon_loss = recon_loss.mean()
        return recon_loss, kld_theta




"""Top-level package for OCTIS."""
from octis.dashboard import *
from octis.configuration import *
from octis.dataset import *
from octis.evaluation_metrics import *
from octis.models import *
from octis.optimization import *
from octis.preprocessing import *

__author__ = """Silvia Terragni"""
__email__ = 's.terragni4@campus.unimib.it'
__version__ = '1.6.0'



import os
import json
from pathlib import Path

# get the path to the framework folder
path = Path(os.path.dirname(os.path.realpath(__file__)))
path = str(path.parent.parent)


def scanDatasets():
    """
    Retrieves the name of each dataset present in the framework

    :return: list with name of each dataset as element
    :rtype: List
    """

    datasets = os.listdir(str(os.path.join(path, "preprocessed_datasets")))
    datasets.remove("README.rst")
    return datasets


def getDatasetMetadata(datasetName):
    """
    Retrieves the dataset metadata

    :param datasetName: name of the dataset
    :type datasetName: String

    :return: dict with metadata if dataset is found, False otherwise
    :rtype: Dict
    """
    file = str(os.path.join(
        path, "preprocessed_datasets", datasetName, "corpus.tsv"))
    if os.path.isfile(file):
        f = open(file,)
        return {"total_documents": sum(1 for line in f)}
    return False


def getDocPreview(datasetName, documentNumber):
    """
    Retrieve the first 40 words of the selected document

    :param datasetName: name of the dataset in which the document is located 
                 (the dataset must be in the preprocessed_datasets folder)
    :type datasetName: String
    :param documentNumber: number of the document to retrieve
    :type documentNumber: Int

    :return: First 40 words in the document
    :rtype: String
    """
    datasetPath = str(os.path.join(
        path, "preprocessed_datasets", datasetName, "corpus.tsv"))
    corpus = []
    if os.path.isfile(datasetPath):
        with open(datasetPath, 'r') as corpus_file:
            for line in corpus_file:
                corpus.append(line.split("\t")[0])
        splitted = corpus[documentNumber].split()
        if len(splitted) > 40:
            return " ".join(splitted[0:40])
        return corpus[documentNumber]
    return False


def getVocabulary(path):
    """
    Retrieves the vocabulary from the vocabulary file of an ezxperiment

    :param path: path of the vocabulary
    :type path: String

    :return: a dictionary with id as a key and word as value,
                 returns False if the vocabulary is not found
    :rtype: Dict
    """
    if Path(path).is_file():
        vocabulary_file = open(path,)
        return json.load(vocabulary_file)
    return False



from octis.models.model import AbstractModel
from gensim.models import lsimodel
import numpy as np
import gensim.corpora as corpora
import octis.configuration.defaults as defaults


class LSI(AbstractModel):

    id2word = None
    id_corpus = None
    hyperparameters = {}
    use_partitions = True
    update_with_test = False

    def __init__(self, num_topics=200, chunksize=20000, decay=1.0,
                 distributed=False, onepass=True, power_iters=2,
                 extra_samples=100):
        """
        Initialize LSI model

        Parameters
        ----------
        num_topics (int, optional) – Number of requested factors

        chunksize (int, optional) – Number of documents to be used in each
        training chunk.

        decay (float, optional) – Weight of existing observations relatively
        to new ones.

        distributed (bool, optional) – If True - distributed mode (parallel
        execution on several machines) will be used.

        onepass (bool, optional) – Whether the one-pass algorithm should be
        used for training. Pass False to force a multi-pass stochastic algorithm.

        power_iters (int, optional) – Number of power iteration steps to be used.
        Increasing the number of power iterations improves accuracy, but lowers
        performance

        extra_samples (int, optional) – Extra samples to be used besides the rank
        k. Can improve accuracy.
        """
        self.hyperparameters["num_topics"] = num_topics
        self.hyperparameters["chunksize"] = chunksize
        self.hyperparameters["decay"] = decay
        self.hyperparameters["distributed"] = distributed
        self.hyperparameters["onepass"] = onepass
        self.hyperparameters["power_iters"] = power_iters
        self.hyperparameters["extra_samples"] = extra_samples

    def info(self):
        """
        Returns model informations
        """
        return {
            "name": "LSI, Latent Semantic Indexing"
        }

    def hyperparameters_info(self):
        """
        Returns hyperparameters informations
        """
        return defaults.LSI_hyperparameters_info

    def partitioning(self, use_partitions, update_with_test=False):
        """
        Handle the partitioning system to use and reset the model to perform
        new evaluations

        Parameters
        ----------
        use_partitions: True if train/set partitioning is needed, False
                        otherwise
        update_with_test: True if the model should be updated with the test set,
                          False otherwise
        """
        self.use_partitions = use_partitions
        self.update_with_test = update_with_test
        self.id2word = None
        self.id_corpus = None

    def train_model(self, dataset, hyperparameters={}, top_words=10):
        """
        Train the model and return output

        Parameters
        ----------
        dataset : dataset to use to build the model
        hyperparameters : hyperparameters to build the model
        top_words : if greather than 0 returns the most significant words
                 for each topic in the output
                 Default True

        Returns
        -------
        result : dictionary with up to 3 entries,
                 'topics', 'topic-word-matrix' and
                 'topic-document-matrix'
        """
        partition = []
        if self.use_partitions:
            partition = dataset.get_partitioned_corpus(use_validation=False)
        else:
            partition = [dataset.get_corpus(), []]

        if self.id2word == None:
            self.id2word = corpora.Dictionary(dataset.get_corpus())

        if self.id_corpus == None:
            self.id_corpus = [self.id2word.doc2bow(
                document) for document in partition[0]]

        hyperparameters["corpus"] = self.id_corpus
        hyperparameters["id2word"] = self.id2word
        self.hyperparameters.update(hyperparameters)

        self.trained_model = lsimodel.LsiModel(**self.hyperparameters)

        result = {}

        result["topic-word-matrix"] = self._get_topic_word_matrix()

        if top_words > 0:
            topics_output = []
            for topic in result["topic-word-matrix"]:
                top_k = np.argsort(topic)[-top_words:]
                top_k_words = list(reversed([self.id2word[i] for i in top_k]))
                topics_output.append(top_k_words)
            result["topics"] = topics_output


        result["topic-document-matrix"] = self._get_topic_document_matrix()

        if self.use_partitions:
            new_corpus = [self.id2word.doc2bow(
                document) for document in partition[1]]
            if self.update_with_test:
                self.trained_model.add_documents(new_corpus)
                self.id_corpus.extend(new_corpus)

                result["test-topic-word-matrix"] = self._get_topic_word_matrix()

                if top_words > 0:
                    topics_output = []
                    for topic in result["test-topic-word-matrix"]:
                        top_k = np.argsort(topic)[-top_words:]
                        top_k_words = list(
                            reversed([self.id2word[i] for i in top_k]))
                        topics_output.append(top_k_words)
                    result["test-topics"] = topics_output

                result["test-topic-document-matrix"] = self._get_topic_document_matrix()

            else:
                test_document_topic_matrix = []
                for document in new_corpus:

                    document_topics_tuples = self.trained_model[document]
                    document_topics = np.zeros(
                        self.hyperparameters["num_topics"])
                    for single_tuple in document_topics_tuples:
                        document_topics[single_tuple[0]] = single_tuple[1]

                    test_document_topic_matrix.append(document_topics)

                result["test-topic-document-matrix"] = np.array(
                    test_document_topic_matrix).transpose()

        return result

    def _get_topic_word_matrix(self):
        """
        Return the topic representation of the words
        """
        topic_word_matrix = self.trained_model.get_topics()
        normalized = []
        for words_w in topic_word_matrix:
            minimum = min(words_w)
            words = words_w - minimum
            normalized.append([float(i)/sum(words) for i in words])
        topic_word_matrix = np.array(normalized)
        return topic_word_matrix

    def _get_topics_words(self, topics):
        """
        Return the most significative words for each topic.
        """
        topic_terms = []
        for i in range(self.hyperparameters["num_topics"]):
            topic_words_list = []
            for word_tuple in self.trained_model.show_topic(i, topics):
                topic_words_list.append(word_tuple[0])
            topic_terms.append(topic_words_list)
        return topic_terms

    def _get_topic_document_matrix(self):
        """
        Return the topic representation of the
        corpus
        """
        topic_weights = self.trained_model[self.id_corpus]

        topic_document = []

        for document_topic_weights in topic_weights:

            # Find min e max topic_weights values
            minimum = document_topic_weights[0][1]
            maximum = document_topic_weights[0][1]
            for topic in document_topic_weights:
                if topic[1] > maximum:
                    maximum = topic[1]
                if topic[1] < minimum:
                    minimum = topic[1]

            # For each topic compute normalized weight
            # in the form (value-min)/(max-min)
            topic_w = []
            for topic in document_topic_weights:
                topic_w.append((topic[1]-minimum)/(maximum-minimum))
            topic_document.append(topic_w)

        return np.array(topic_document).transpose()



from octis.models.model import AbstractModel
import numpy as np
from gensim.models import nmf
import gensim.corpora as corpora
import octis.configuration.citations as citations
import octis.configuration.defaults as defaults


class NMF(AbstractModel):

    def __init__(self, num_topics=100, chunksize=2000, passes=1, kappa=1.0, minimum_probability=0.01, w_max_iter=200,
                 w_stop_condition=0.0001, h_max_iter=50, h_stop_condition=0.001, eval_every=10, normalize=True,
                 random_state=None, use_partitions=True):
        """
        Initialize NMF model

        Parameters
        ----------
        num_topics (int, optional) – Number of topics to extract.

        chunksize (int, optional) – Number of documents to be used in each
        training chunk.

        passes (int, optional) – Number of full passes over the
        training corpus. Leave at default passes=1 if your input
        is an iterator.

        kappa (float, optional) – Gradient descent step size.
        Larger value makes the model train faster, but could
        lead to non-convergence if set too large.

        minimum_probability – If normalize is True, topics with
        smaller probabilities are filtered out. If normalize is False,
        topics with smaller factors are filtered out. If set to None,
        a value of 1e-8 is used to prevent 0s.

        w_max_iter (int, optional) – Maximum number of iterations to
        train W per each batch.

        w_stop_condition (float, optional) – If error difference gets less
        than that, training of W stops for the current batch.

        h_max_iter (int, optional) – Maximum number of iterations to train
        h per each batch.

        h_stop_condition (float) – If error difference gets less than that,
        training of h stops for the current batch.

        eval_every (int, optional) – Number of batches after which l2 norm
        of (v - Wh) is computed. Decreases performance if set too low.

        normalize (bool or None, optional) – Whether to normalize the result.

        random_state ({np.random.RandomState, int}, optional) – Seed for
        random generator. Needed for reproducibility.
        """
        super().__init__()
        self.hyperparameters["num_topics"] = num_topics
        self.hyperparameters["chunksize"] = chunksize
        self.hyperparameters["passes"] = passes
        self.hyperparameters["kappa"] = kappa
        self.hyperparameters["minimum_probability"] = minimum_probability
        self.hyperparameters["w_max_iter"] = w_max_iter
        self.hyperparameters["w_stop_condition"] = w_stop_condition
        self.hyperparameters["h_max_iter"] = h_max_iter
        self.hyperparameters["h_stop_condition"] = h_stop_condition
        self.hyperparameters["eval_every"] = eval_every
        self.hyperparameters["normalize"] = normalize
        self.hyperparameters["random_state"] = random_state
        self.use_partitions = use_partitions

        self.id2word = None
        self.id_corpus = None
        self.update_with_test = False

    def info(self):
        """
        Returns model informations
        """
        return {
            "citation": citations.models_NMF,
            "name": "NMF, Non-negative Matrix Factorization"
        }

    def hyperparameters_info(self):
        """
        Returns hyperparameters informations
        """
        return defaults.NMF_gensim_hyperparameters_info

    def partitioning(self, use_partitions, update_with_test=False):
        """
        Handle the partitioning system to use and reset the model to perform
        new evaluations

        Parameters
        ----------
        use_partitions: True if train/set partitioning is needed, False
                        otherwise
        update_with_test: True if the model should be updated with the test set,
                          False otherwise
        """
        self.use_partitions = use_partitions
        self.update_with_test = update_with_test
        self.id2word = None
        self.id_corpus = None

    def train_model(self, dataset, hyperparameters=None, top_words=10):
        """
        Train the model and return output

        Parameters
        ----------
        dataset : dataset to use to build the model
        hyperparameters : hyperparameters to build the model
        top_words : if greather than 0 returns the most significant words
                 for each topic in the output
                 Default True

        Returns
        -------
        result : dictionary with up to 3 entries,
                 'topics', 'topic-word-matrix' and
                 'topic-document-matrix'
        """
        if hyperparameters is None:
            hyperparameters = {}
        if self.use_partitions:
            partition = dataset.get_partitioned_corpus(use_validation=False)
        else:
            partition = [dataset.get_corpus(), []]

        if self.id2word is None:
            self.id2word = corpora.Dictionary(dataset.get_corpus())
        if self.id_corpus is None:
            self.id_corpus = [self.id2word.doc2bow(
                document) for document in partition[0]]

        hyperparameters["corpus"] = self.id_corpus
        hyperparameters["id2word"] = self.id2word
        self.hyperparameters.update(hyperparameters)

        self.trained_model = nmf.Nmf(**self.hyperparameters)

        result = {}

        result["topic-word-matrix"] = self.trained_model.get_topics()

        if top_words > 0:
            topics_output = []
            for topic in result["topic-word-matrix"]:
                top_k = np.argsort(topic)[-top_words:]
                top_k_words = list(reversed([self.id2word[i] for i in top_k]))
                topics_output.append(top_k_words)
            result["topics"] = topics_output

        result["topic-document-matrix"] = self._get_topic_document_matrix()

        if self.use_partitions:
            new_corpus = [self.id2word.doc2bow(
                document) for document in partition[1]]
            if self.update_with_test:
                self.trained_model.update(new_corpus)
                self.id_corpus.extend(new_corpus)

                result["test-topic-word-matrix"] = self.trained_model.get_topics()

                if top_words > 0:
                    topics_output = []
                    for topic in result["test-topic-word-matrix"]:
                        top_k = np.argsort(topic)[-top_words:]
                        top_k_words = list(
                            reversed([self.id2word[i] for i in top_k]))
                        topics_output.append(top_k_words)
                    result["test-topics"] = topics_output

                result["test-topic-document-matrix"] = self._get_topic_document_matrix()
            else:
                result["test-topic-document-matrix"] = self._get_topic_document_matrix(new_corpus)
        return result

    def _get_topics_words(self, topk):
        """
        Return the most significative words for each topic.
        """
        topic_terms = []
        for i in range(self.hyperparameters["num_topics"]):
            topic_words_list = []
            for word_tuple in self.trained_model.get_topic_terms(i, topk):
                topic_words_list.append(self.id2word[word_tuple[0]])
            topic_terms.append(topic_words_list)
        return topic_terms

    def _get_topic_document_matrix(self, test_corpus=None):
        """
        Return the topic representation of the
        corpus
        """
        doc_topic_tuples = []

        if test_corpus is None:
            for document in self.id_corpus:
                doc_topic_tuples.append(
                    self.trained_model.get_document_topics(document, minimum_probability=0))
        else:
            for document in test_corpus:
                doc_topic_tuples.append(
                    self.trained_model.get_document_topics(document, minimum_probability=0))
        topic_document = np.zeros((
            self.hyperparameters["num_topics"],
            len(doc_topic_tuples)))

        for ndoc in range(len(doc_topic_tuples)):
            document = doc_topic_tuples[ndoc]
            for topic_tuple in document:
                topic_document[topic_tuple[0]][ndoc] = topic_tuple[1]
        return topic_document



from __future__ import print_function
from octis.models.early_stopping.pytorchtools import EarlyStopping

import torch
import numpy as np
from octis.models.ETM_model import data
from sklearn.feature_extraction.text import CountVectorizer
from torch import nn, optim
from octis.models.ETM_model import etm
from octis.models.model import AbstractModel
import gensim
import pickle as pkl


class ETM(AbstractModel):

    def __init__(self, num_topics=10, num_epochs=100, t_hidden_size=800, rho_size=300, embedding_size=300,
                 activation='relu', dropout=0.5, lr=0.005, optimizer='adam', batch_size=128, clip=0.0,
                 wdecay=1.2e-6, bow_norm=1, device='cpu', top_word=10, train_embeddings=True, embeddings_path=None,
                 use_partitions=True):
        super(ETM, self).__init__()
        self.hyperparameters = dict()
        self.hyperparameters['num_topics'] = int(num_topics)
        self.hyperparameters['num_epochs'] = int(num_epochs)
        self.hyperparameters['t_hidden_size'] = int(t_hidden_size)
        self.hyperparameters['rho_size'] = int(rho_size)
        self.hyperparameters['embedding_size'] = int(embedding_size)
        self.hyperparameters['activation'] = activation
        self.hyperparameters['dropout'] = float(dropout)
        self.hyperparameters['lr'] = float(lr)
        self.hyperparameters['optimizer'] = optimizer
        self.hyperparameters['batch_size'] = int(batch_size)
        self.hyperparameters['clip'] = float(clip)
        self.hyperparameters['wdecay'] = float(wdecay)
        self.hyperparameters['bow_norm'] = int(bow_norm)
        self.hyperparameters['train_embeddings'] = bool(train_embeddings)
        self.hyperparameters['embeddings_path'] = embeddings_path
        self.top_word = top_word
        self.early_stopping = None
        self.device = device
        self.test_tokens, self.test_counts = None, None
        self.valid_tokens, self.valid_counts = None, None
        self.train_tokens, self.train_counts, self.vocab = None, None, None
        self.use_partitions = use_partitions
        self.model = None
        self.optimizer = None
        self.embeddings = None

    def train_model(self, dataset, hyperparameters=None, top_words=10):
        if hyperparameters is None:
            hyperparameters = {}
        self.set_model(dataset, hyperparameters)
        self.top_word = top_words
        self.early_stopping = EarlyStopping(patience=5, verbose=True)

        for epoch in range(0, self.hyperparameters['num_epochs']):
            continue_training = self._train_epoch(epoch)
            if not continue_training:
                break

        # load the last checkpoint with the best model
        # self.model.load_state_dict(torch.load('etm_checkpoint.pt'))

        if self.use_partitions:
            result = self.inference()
        else:
            result = self.get_info()

        return result

    def set_model(self, dataset, hyperparameters):
        if self.use_partitions:
            train_data, validation_data, testing_data = dataset.get_partitioned_corpus(use_validation=True)

            data_corpus_train = [' '.join(i) for i in train_data]
            data_corpus_test = [' '.join(i) for i in testing_data]
            data_corpus_val = [' '.join(i) for i in validation_data]

            vocab = dataset.get_vocabulary()
            self.vocab = {i: w for i, w in enumerate(vocab)}
            vocab2id = {w: i for i, w in enumerate(vocab)}

            self.train_tokens, self.train_counts, self.test_tokens, self.test_counts, self.valid_tokens, \
            self.valid_counts = self.preprocess(vocab2id, data_corpus_train, data_corpus_test, data_corpus_val)

        else:
            data_corpus = [' '.join(i) for i in dataset.get_corpus()]
            vocab = dataset.get_vocabulary()
            self.vocab = {i: w for i, w in enumerate(vocab)}
            vocab2id = {w: i for i, w in enumerate(vocab)}

            self.train_tokens, self.train_counts = self.preprocess(vocab2id, data_corpus, None)

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self.set_default_hyperparameters(hyperparameters)
        self.load_embeddings()
        ## define model and optimizer
        self.model = etm.ETM(num_topics=self.hyperparameters['num_topics'], vocab_size=len(self.vocab.keys()),
                             t_hidden_size=int(self.hyperparameters['t_hidden_size']),
                             rho_size=int(self.hyperparameters['rho_size']),
                             emb_size=int(self.hyperparameters['embedding_size']),
                             theta_act=self.hyperparameters['activation'],
                             embeddings=self.embeddings,
                             train_embeddings=self.hyperparameters['train_embeddings'],
                             enc_drop=self.hyperparameters['dropout']).to(self.device)
        print('model: {}'.format(self.model))

        self.optimizer = self.set_optimizer()

    def set_optimizer(self):
        self.hyperparameters['lr'] = float(self.hyperparameters['lr'])
        self.hyperparameters['wdecay'] = float(self.hyperparameters['wdecay'])
        if self.hyperparameters['optimizer'] == 'adam':
            optimizer = optim.Adam(self.model.parameters(), lr=self.hyperparameters['lr'],
                                   weight_decay=self.hyperparameters['wdecay'])
        elif self.hyperparameters['optimizer'] == 'adagrad':
            optimizer = optim.Adagrad(self.model.parameters(), lr=self.hyperparameters['lr'],
                                      weight_decay=self.hyperparameters['wdecay'])
        elif self.hyperparameters['optimizer'] == 'adadelta':
            optimizer = optim.Adadelta(self.model.parameters(), lr=self.hyperparameters['lr'],
                                       weight_decay=self.hyperparameters['wdecay'])
        elif self.hyperparameters['optimizer'] == 'rmsprop':
            optimizer = optim.RMSprop(self.model.parameters(), lr=self.hyperparameters['lr'],
                                      weight_decay=self.hyperparameters['wdecay'])
        elif self.hyperparameters['optimizer'] == 'asgd':
            optimizer = optim.ASGD(self.model.parameters(), lr=self.hyperparameters['lr'],
                                   t0=0, lambd=0., weight_decay=self.hyperparameters['wdecay'])
        elif self.hyperparameters['optimizer'] == 'sgd':
            optimizer = optim.SGD(self.model.parameters(), lr=self.hyperparameters['lr'],
                                  weight_decay=self.hyperparameters['wdecay'])
        else:
            print('Defaulting to vanilla SGD')
            optimizer = optim.SGD(self.model.parameters(), lr=self.hyperparameters['lr'])

        return optimizer

    def _train_epoch(self, epoch):
        self.data_list = []
        self.model.train()
        acc_loss = 0
        acc_kl_theta_loss = 0
        cnt = 0
        indices = torch.arange(0, len(self.train_tokens))
        indices = torch.split(indices, self.hyperparameters['batch_size'])
        for idx, ind in enumerate(indices):
            self.optimizer.zero_grad()
            self.model.zero_grad()
            data_batch = data.get_batch(self.train_tokens, self.train_counts, ind, len(self.vocab.keys()),
                                        self.hyperparameters['embedding_size'], self.device)
            sums = data_batch.sum(1).unsqueeze(1)
            if self.hyperparameters['bow_norm']:
                normalized_data_batch = data_batch / sums
            else:
                normalized_data_batch = data_batch
            recon_loss, kld_theta = self.model(data_batch, normalized_data_batch)
            total_loss = recon_loss + kld_theta
            total_loss.backward()
            if self.hyperparameters["clip"] > 0:
                torch.nn.utils.clip_grad_norm_(self.model.parameters(),
                                               self.hyperparameters["clip"])
            self.optimizer.step()

            acc_loss += torch.sum(recon_loss).item()
            acc_kl_theta_loss += torch.sum(kld_theta).item()
            cnt += 1
            log_interval = 20
            if idx % log_interval == 0 and idx > 0:
                cur_loss = round(acc_loss / cnt, 2)
                cur_kl_theta = round(acc_kl_theta_loss / cnt, 2)
                cur_real_loss = round(cur_loss + cur_kl_theta, 2)

                print('Epoch: {} .. batch: {}/{} .. LR: {} .. KL_theta: {} .. Rec_loss: {}'
                      ' .. NELBO: {}'.format(epoch + 1, idx, len(indices),
                                             self.optimizer.param_groups[0]['lr'],
                                             cur_kl_theta, cur_loss, cur_real_loss))

            self.data_list.append(normalized_data_batch)

        cur_loss = round(acc_loss / cnt, 2)
        cur_kl_theta = round(acc_kl_theta_loss / cnt, 2)
        cur_real_loss = round(cur_loss + cur_kl_theta, 2)
        print('*' * 100)
        print('Epoch----->{} .. LR: {} .. KL_theta: {} .. Rec_loss: {} .. NELBO: {}'.format(
            epoch + 1, self.optimizer.param_groups[0]['lr'], cur_kl_theta, cur_loss,
            cur_real_loss))
        print('*' * 100)

        # VALIDATION ###
        if self.valid_tokens is None:
            return True
        else:
            model = self.model.to(self.device)
            model.eval()
            with torch.no_grad():
                val_acc_loss = 0
                val_acc_kl_theta_loss = 0
                val_cnt = 0
                indices = torch.arange(0, len(self.valid_tokens))
                indices = torch.split(indices, self.hyperparameters['batch_size'])
                for idx, ind in enumerate(indices):
                    self.optimizer.zero_grad()
                    self.model.zero_grad()
                    val_data_batch = data.get_batch(self.valid_tokens, self.valid_counts,
                                                    ind, len(self.vocab.keys()),
                                                    self.hyperparameters['embedding_size'], self.device)
                    sums = val_data_batch.sum(1).unsqueeze(1)
                    if self.hyperparameters['bow_norm']:
                        val_normalized_data_batch = val_data_batch / sums
                    else:
                        val_normalized_data_batch = val_data_batch

                    val_recon_loss, val_kld_theta = self.model(val_data_batch,
                                                               val_normalized_data_batch)

                    val_acc_loss += torch.sum(val_recon_loss).item()
                    val_acc_kl_theta_loss += torch.sum(val_kld_theta).item()
                    val_cnt += 1
                    val_total_loss = val_recon_loss + val_kld_theta

                val_cur_loss = round(val_acc_loss / cnt, 2)
                val_cur_kl_theta = round(val_acc_kl_theta_loss / cnt, 2)
                val_cur_real_loss = round(val_cur_loss + val_cur_kl_theta, 2)
                print('*' * 100)
                print('VALIDATION .. LR: {} .. KL_theta: {} .. Rec_loss: {} .. NELBO: {}'.format(
                    self.optimizer.param_groups[0]['lr'], val_cur_kl_theta, val_cur_loss,
                    val_cur_real_loss))
                print('*' * 100)
                if np.isnan(val_cur_real_loss):
                    return False
                else:
                    self.early_stopping(val_total_loss, model)

                    if self.early_stopping.early_stop:
                        print("Early stopping")
                        return False
                    else:
                        return True

    def get_info(self):
        topic_w = []
        self.model.eval()
        info = {}
        with torch.no_grad():
            theta, _ = self.model.get_theta(torch.cat(self.data_list))
            gammas = self.model.get_beta().cpu().numpy()
            for k in range(self.hyperparameters['num_topics']):
                if np.isnan(gammas[k]).any():
                    # to deal with nan matrices
                    topic_w = None
                    break
                else:
                    top_words = list(gammas[k].argsort()[-self.top_word:][::-1])
                topic_words = [self.vocab[a] for a in top_words]
                topic_w.append(topic_words)

        info['topic-word-matrix'] = gammas
        info['topic-document-matrix'] = theta.cpu().detach().numpy().T
        info['topics'] = topic_w
        print(info['topics'])
        return info

    def inference(self):
        assert isinstance(self.use_partitions, bool) and self.use_partitions
        topic_d = []
        self.model.eval()
        indices = torch.arange(0, len(self.test_tokens))
        indices = torch.split(indices, self.hyperparameters['batch_size'])

        for idx, ind in enumerate(indices):
            data_batch = data.get_batch(self.test_tokens, self.test_counts,
                                        ind, len(self.vocab.keys()),
                                        self.hyperparameters['embedding_size'], self.device)
            sums = data_batch.sum(1).unsqueeze(1)
            if self.hyperparameters['bow_norm']:
                normalized_data_batch = data_batch / sums
            else:
                normalized_data_batch = data_batch
            theta, _ = self.model.get_theta(normalized_data_batch)
            topic_d.append(theta.cpu().detach().numpy())

        info = self.get_info()
        emp_array = np.empty((0, self.hyperparameters['num_topics']))
        topic_doc = np.asarray(topic_d)
        length = topic_doc.shape[0]
        # batch concatenation
        for i in range(length):
            emp_array = np.concatenate([emp_array, topic_doc[i]])
        info['test-topic-document-matrix'] = emp_array.T

        return info

    def set_default_hyperparameters(self, hyperparameters):
        for k in hyperparameters.keys():
            if k in self.hyperparameters.keys():
                self.hyperparameters[k] = hyperparameters.get(k, self.hyperparameters[k])

    def partitioning(self, use_partitions=False):
        self.use_partitions = use_partitions

    @staticmethod
    def preprocess(vocab2id, train_corpus, test_corpus=None, validation_corpus=None):
        # def split_bow(bow_in, n_docs):
        #    indices = np.asarray([np.asarray([w for w in bow_in[doc, :].indices]) for doc in range(n_docs)])
        #    counts = np.asarray([np.asarray([c for c in bow_in[doc, :].data]) for doc in range(n_docs)])
        #    return np.expand_dims(indices, axis=0), np.expand_dims(counts, axis=0)

        def split_bow(bow_in, n_docs):
            indices = [[w for w in bow_in[doc, :].indices] for doc in range(n_docs)]
            counts = [[c for c in bow_in[doc, :].data] for doc in range(n_docs)]
            return indices, counts

        vec = CountVectorizer(
            vocabulary=vocab2id, token_pattern=r'(?u)\b\w+\b')

        dataset = train_corpus.copy()
        if test_corpus is not None:
            dataset.extend(test_corpus)
        if validation_corpus is not None:
            dataset.extend(validation_corpus)

        vec.fit(dataset)
        idx2token = {v: k for (k, v) in vec.vocabulary_.items()}

        x_train = vec.transform(train_corpus)
        x_train_tokens, x_train_count = split_bow(x_train, x_train.shape[0])

        if test_corpus is not None:
            x_test = vec.transform(test_corpus)
            x_test_tokens, x_test_count = split_bow(x_test, x_test.shape[0])

            if validation_corpus is not None:
                x_validation = vec.transform(validation_corpus)
                x_val_tokens, x_val_count = split_bow(x_validation, x_validation.shape[0])

                return x_train_tokens, x_train_count, x_test_tokens, x_test_count, x_val_tokens, x_val_count
            else:
                return x_train_tokens, x_train_count, x_test_tokens, x_test_count
        else:
            if validation_corpus is not None:
                x_validation = vec.transform(validation_corpus)
                x_val_tokens, x_val_count = split_bow(x_validation, x_validation.shape[0])
                return x_train_tokens, x_train_count, x_val_tokens, x_val_count
            else:
                return x_train_tokens, x_train_count

    def load_embeddings(self):
        if not self.hyperparameters['train_embeddings']:
            vectors = {}
            embs = pkl.load(open(self.hyperparameters['embeddings_path'], 'rb'))
            for l in embs:
                line = l.split()
                word = line[0]
                if word in self.vocab.values():
                    vect = np.array(line[1:]).astype(np.float)
                    vectors[word] = vect
            embeddings = np.zeros((len(self.vocab.keys()), self.hyperparameters['embedding_size']))
            words_found = 0
            for i, word in enumerate(self.vocab.values()):
                try:
                    embeddings[i] = vectors[word]
                    words_found += 1
                except KeyError:
                    embeddings[i] = np.random.normal(scale=0.6, size=(self.hyperparameters['embedding_size'],))
            self.embeddings = torch.from_numpy(embeddings).to(self.device)

    def filter_pretrained_embeddings(self, pretrained_embeddings_path, save_embedding_path, vocab_path, binary=True):
        """
        Filter the embeddings from a set of word2vec-format pretrained embeddings based on the vocabulary
        This should allow you to avoid to load the whole embedding space every time you do Bayesian Optimization
        but just the embeddings that are in the vocabulary.
        :param pretrained_embeddings_path:
        :return:
        """
        vocab = []
        with open(vocab_path, 'r') as fr:
            for line in fr.readlines():
                vocab.append(line.strip().split(" ")[0])

        w2v_model = gensim.models.KeyedVectors.load_word2vec_format(pretrained_embeddings_path, binary=binary)
        embeddings = []
        for word in vocab:
            if word in w2v_model.vocab:
                line = word
                for w in w2v_model[word].tolist():
                    line = line + " " + str(w)
                embeddings.append(line)
        pkl.dump(embeddings, open(save_embedding_path, 'wb'))



from os import environ, makedirs
from os.path import exists, expanduser, join, splitext
import pickle
import sys
import codecs
import shutil
import requests
import json

"""
This code is highly inspired by the scikit-learn strategy to download datasets
"""


def get_data_home(data_home=None):
    """Return the path of the octis data dir.
    By default the data dir is set to a folder named 'octis_data' in the
    user home folder.
    Alternatively, it can be set by the 'OCTIS_DATA' environment
    variable or programmatically by giving an explicit folder path. The '~'
    symbol is expanded to the user home folder.
    If the folder does not already exist, it is automatically created.
    Parameters
    ----------
    data_home : str | None
        The path to octis data dir.
    """
    if data_home is None:
        data_home = environ.get('OCTIS_DATA', join('~', 'octis_data'))
    data_home = expanduser(data_home)
    if not exists(data_home):
        makedirs(data_home)
    return data_home


def _pkl_filepath(*args, **kwargs):
    """Ensure different filenames for Python 2 and Python 3 pickles
    An object pickled under Python 3 cannot be loaded under Python 2. An object
    pickled under Python 2 can sometimes not be loaded correctly under Python 3
    because some Python 2 strings are decoded as Python 3 strings which can be
    problematic for objects that use Python 2 strings as byte buffers for
    numerical data instead of "real" strings.
    Therefore, dataset loaders in octis use different files for pickles
    manages by Python 2 and Python 3 in the same OCTIS_DATA folder so as
    to avoid conflicts.
    args[-1] is expected to be the ".pkl" filename. Under Python 3, a suffix is
    inserted before the extension to s
    _pkl_filepath('/path/to/folder', 'filename.pkl') returns:
      - /path/to/folder/filename.pkl under Python 2
      - /path/to/folder/filename_py3.pkl under Python 3+
    """
    py3_suffix = kwargs.get("py3_suffix", "_py3")
    basename, ext = splitext(args[-1])
    if sys.version_info[0] >= 3:
        basename += py3_suffix
    new_args = args[:-1] + (basename + ext,)
    return join(*new_args)


def download_dataset(dataset_name, target_dir, cache_path):
    """Download the 20 newsgroups data and stored it as a zipped pickle."""
    corpus_path = join(target_dir, "corpus.tsv")
    metadata_path = join(target_dir, "metadata.json")
    vocabulary_path = join(target_dir, "vocabulary.txt")

    if not exists(target_dir):
        makedirs(target_dir)

    dataset_url = "https://raw.githubusercontent.com/MIND-Lab/OCTIS/master/preprocessed_datasets/" + dataset_name

    corpus = requests.get(dataset_url + "/corpus.tsv")
    metadata = requests.get(dataset_url + "/metadata.json")
    vocabulary = requests.get(dataset_url + "/vocabulary.txt")

    if corpus and metadata and vocabulary:
        with open(corpus_path, 'w') as f:
            f.write(corpus.text)
        with open(metadata_path, 'w') as f:
            f.write(metadata.text)
        with open(vocabulary_path, 'w') as f:
            f.write(vocabulary.text)

        only_docs, labels, partition = [], [], []
        for d in corpus.text.split("\n"):
            if len(d.strip()) > 0:
                dsplit = d.strip().split("\t")
                only_docs.append(dsplit[0])
                if len(dsplit) > 1:
                    partition.append(dsplit[1])
                    if len(dsplit) > 2:
                        labels.append(dsplit[2])

        vocab = [word for word in vocabulary.text.split("\n") if len(word) > 0]
        metadata = json.loads(metadata.text)

        metadata["info"]["name"] = dataset_name

        # Store a zipped pickle
        cache = dict(corpus=only_docs, labels=labels, partitions=partition, metadata=metadata,
                     vocabulary=vocab)
        compressed_content = codecs.encode(pickle.dumps(cache), 'zlib_codec')
        with open(cache_path, 'wb') as f:
            f.write(compressed_content)

        shutil.rmtree(target_dir)
        return cache
    else:
        raise Exception(dataset_name + ' dataset not found')





import numpy as np
import octis.configuration.citations as citations
from octis.evaluation_metrics.metrics import AbstractMetric


def _KL(P, Q):
    """
    Perform Kullback-Leibler divergence

    Parameters
    ----------
    P : distribution P
    Q : distribution Q

    Returns
    -------
    divergence : divergence from Q to P
    """
    # add epsilon to grant absolute continuity
    epsilon = 0.00001
    P = P+epsilon
    Q = Q+epsilon

    divergence = np.sum(P*np.log(P/Q))
    return divergence


def _replace_zeros_lines(arr):
    zero_lines = np.where(~arr.any(axis=1))[0]
    val = 1.0 / len(arr[0])
    vett = np.full(len(arr[0]), val)
    for zero_line in zero_lines:
        arr[zero_line] = vett.copy()
    return arr


class KL_uniform(AbstractMetric):
    def __init__(self):
        """
        Initialize metric
        """
        super().__init__()

    def info(self):
        return {
            "citation": citations.em_topic_significance,
            "name": "KL_Uniform, Uniform distribution over words"
        }

    def score(self, model_output):
        """
        Retrieves the score of the metric

        Parameters
        ----------
        model_output : dictionary, output of the model
                       'topic-word-matrix' required

        Returns
        -------
        result : score
        """
        phi = _replace_zeros_lines(model_output["topic-word-matrix"].astype(float))

        # make uniform distribution
        val = 1.0 / len(phi[0])
        unif_distr = np.full(len(phi[0]), val)

        divergences = []
        for topic in range(len(phi)):

            # normalize phi, sum up to 1
            P = phi[topic] / phi[topic].sum()

            divergence = _KL(P, unif_distr)
            divergences.append(divergence)

        # KL-uniform = mean of the divergences
        # between topic-word distributions and uniform distribution
        result = np.array(divergences).mean()
        return result


class KL_vacuous(AbstractMetric):
    def __init__(self):
        """
        Initialize metric
        """
        super().__init__()

    def info(self):
        return {
            "citation": citations.em_topic_significance,
            "name": "KL_Vacuous, Vacuous semantic distribution"
        }

    def score(self, model_output):
        """
        Retrieves the score of the metric

        Parameters
        ----------
        model_output : dictionary, output of the model
                       'topic-word-matrix' required
                       'topic-document-matrix' required

        Returns
        -------
        result : score
        """
        phi = _replace_zeros_lines(model_output["topic-word-matrix"].astype(float))
        theta = _replace_zeros_lines(model_output["topic-document-matrix"].astype(float))

        vacuous = np.zeros(phi.shape[1])
        for topic in range(len(theta)):

            # get probability of the topic in the corpus
            p_topic = theta[topic].sum()/len(theta[0])

            # get probability of the words:
            # P(Wi | vacuous_dist) = P(Wi | topic)*P(topic)
            vacuous += phi[topic]*p_topic

        divergences = []
        for topic in range(len(phi)):

            # normalize phi, sum up to 1
            P = phi[topic] / phi[topic].sum()

            divergence = _KL(P, vacuous)
            divergences.append(divergence)

        # KL-vacuous = mean of the divergences between topic-word distributions and vacuous distribution
        result = np.array(divergences).mean()
        return result


class KL_background(AbstractMetric):
    def __init__(self):
        """
        Initialize metric
        """
        super().__init__()

    def info(self):
        return {
            "citation": citations.em_topic_significance,
            "name": "KL_Background, Background distribution over documents"
        }

    def score(self, model_output):
        """
        Retrieves the score of the metric

        Parameters
        ----------
        model_output : dictionary, output of the model
                       'topic-document-matrix' required

        Returns
        -------
        result : score
        """
        theta = _replace_zeros_lines(model_output["topic-document-matrix"].astype(float))

        # make uniform distribution
        val = 1.0 / len(theta[0])
        unif_distr = np.full(len(theta[0]), val)

        divergences = []
        for topic in range(len(theta)):
            # normalize theta, sum up to 1
            P = theta[topic] / theta[topic].sum()

            divergence = _KL(P, unif_distr)
            divergences.append(divergence)

        # KL-background = mean of the divergences
        # between topic-doc distributions and uniform distribution
        result = np.array(divergences).mean()
        if np.isnan(result):
            return 0
        return result



from octis.evaluation_metrics.metrics import AbstractMetric
from octis.dataset.dataset import Dataset
from gensim.corpora.dictionary import Dictionary
from gensim.models import CoherenceModel
from gensim.models import KeyedVectors
import gensim.downloader as api
import octis.configuration.citations as citations
import numpy as np
import itertools
from scipy import spatial
from sklearn.preprocessing import normalize
from operator import add


class Coherence(AbstractMetric):
    def __init__(self, texts=None, topk=10, measure='c_npmi'):
        """
        Initialize metric

        Parameters
        ----------
        texts : list of documents (list of lists of strings)
        topk : how many most likely words to consider in
        the evaluation
        measure : (default 'c_npmi') measure to use.
        other measures: 'u_mass', 'c_v', 'c_uci', 'c_npmi'
        """
        super().__init__()
        if texts is None:
            self._texts = _load_default_texts()
        else:
            self._texts = texts
        self._dictionary = Dictionary(self._texts)
        self.topk = topk
        self.measure = measure

    def info(self):
        return {
            "citation": citations.em_coherence,
            "name": "Coherence"
        }

    def score(self, model_output):
        """
        Retrieve the score of the metric

        Parameters
        ----------
        model_output : dictionary, output of the model
                       key 'topics' required.

        Returns
        -------
        score : coherence score
        """
        topics = model_output["topics"]
        if topics is None:
            return -1
        if self.topk > len(topics[0]):
            raise Exception('Words in topics are less than topk')
        else:
            npmi = CoherenceModel(topics=topics, texts=self._texts, dictionary=self._dictionary,
                                  coherence=self.measure, processes=1, topn=self.topk)
            return npmi.get_coherence()


class WECoherencePairwise(AbstractMetric):
    def __init__(self, word2vec_path=None, binary=False, topk=10):
        """
        Initialize metric

        Parameters
        ----------
        dictionary with keys
        topk : how many most likely words to consider
        word2vec_path : if word2vec_file is specified retrieves word embeddings file (in word2vec format)
        to compute similarities, otherwise 'word2vec-google-news-300' is downloaded
        binary : True if the word2vec file is binary, False otherwise (default False)
        """
        super().__init__()

        self.binary = binary
        self.topk = topk
        self.word2vec_path = word2vec_path
        if word2vec_path is None:
            self._wv = api.load('word2vec-google-news-300')
        else:
            self._wv = KeyedVectors.load_word2vec_format(
                word2vec_path, binary=self.binary)

    def info(self):
        return {
            "citation": citations.em_coherence_we,
            "name": "Coherence word embeddings pairwise cosine"
        }

    def score(self, model_output):
        """
        Retrieve the score of the metric

        Parameters
        ----------
        model_output : dictionary, output of the model
                       key 'topics' required.

        Returns
        -------
        score : topic coherence computed on the word embeddings
                similarities
        """
        topics = model_output["topics"]
        if self.topk > len(topics[0]):
            raise Exception('Words in topics are less than topk')
        else:
            arrays = []
            for _, topic in enumerate(topics):
                if len(topic) > 0:
                    local_simi = []
                    for w1, w2 in itertools.combinations(topic[0:self.topk], 2):
                        if w1 in self._wv.vocab and w2 in self._wv.vocab:
                            local_simi.append(self._wv.similarity(w1, w2))
                    arrays.append(np.mean(local_simi))
            return np.mean(arrays)


class WECoherenceCentroid(AbstractMetric):
    def __init__(self, topk=10, word2vec_path=None, binary=True):
        """
        Initialize metric

        Parameters
        ----------
        topk : how many most likely words to consider
        w2v_model_path : a word2vector model path, if not provided, google news 300 will be used instead
        """
        super().__init__()

        self.topk = topk
        self.binary = binary
        self.word2vec_path = word2vec_path
        if self.word2vec_path is None:
            self._wv = api.load('word2vec-google-news-300')
        else:
            self._wv = KeyedVectors.load_word2vec_format(self.word2vec_path, binary=self.binary)

    @staticmethod
    def info():
        return {
            "citation": citations.em_word_embeddings_pc,
            "name": "Coherence word embeddings centroid"
        }

    def score(self, model_output):
        """
        Retrieve the score of the metric

        :param model_output: dictionary, output of the model. key 'topics' required.
        :return topic coherence computed on the word embeddings

        """
        topics = model_output["topics"]
        if self.topk > len(topics[0]):
            raise Exception('Words in topics are less than topk')
        else:
            result = 0
            count = 0
            for topic in topics:
                topic_coherence = 0
                for w1, w2 in itertools.combinations(topic, 2):
                    if w1 in self._wv.vocab and w2 in self._wv.vocab:
                        distance = spatial.distance.cosine(self._wv.__getitem__(w1), self._wv.__getitem__(w2))
                        topic_coherence += distance - 1
                        count = count + 1
                topic_coherence = topic_coherence/count
                result += topic_coherence
            result /= len(topics)
            return result


def _load_default_texts():
    """
    Loads default general texts

    Returns
    -------
    result : default 20newsgroup texts
    """
    dataset = Dataset()
    dataset.fetch_dataset("20NewsGroup")
    return dataset.get_corpus()



import matplotlib.pyplot as plt
import numpy as np
from skopt.learning import GaussianProcessRegressor, RandomForestRegressor, ExtraTreesRegressor
from skopt import Optimizer as skopt_optimizer
from skopt.utils import dimensions_aslist
import os
import importlib
import sys
import octis.configuration.defaults as defaults
from pathlib import Path

framework_path = Path(os.path.dirname(os.path.realpath(__file__)))
framework_path = str(framework_path.parent)


def importClass(class_name, module_name, module_path):
    """
    Import a class runtime based on its module and name

    :param class_name: name of the class
    :type class_name: str
    :param module_name: name of the module
    :type module_name: str
    :param module_path: absolute path to the module
    :type module_path: str
    :return: class object
    :rtype: class
    """
    spec = importlib.util.spec_from_file_location(
        module_name, module_path, submodule_search_locations=[])
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    importlib.invalidate_caches()
    imported_class = getattr(module, class_name)

    return imported_class


def load_model(optimization_object):
    """
    Load the topic model for the resume of the optimization

    :param optimization_object: dictionary of optimization attributes saved in the jaon file
    :type optimization_object: dict
    :return: topic model used during the BO.
    :rtype: object model
    """

    model_parameters = optimization_object['model_attributes']
    use_partitioning = optimization_object['use_partitioning']

    model_name = optimization_object['model_name']
    module_path = os.path.join(framework_path, "models")
    module_path = os.path.join(module_path, model_name + ".py")
    model = importClass(model_name, model_name, module_path)
    model_instance = model()
    model_instance.hyperparameters.update(model_parameters)
    model_instance.use_partitions = use_partitioning

    return model_instance


def select_metric(metric_parameters, metric_name):
    """
    Select the metric for the resume of the optimization

    :param metric_parameters: metric parameters
    :type metric_parameters: list
    :param metric_name: name of the metric
    :type metric_name: str
    :return: metric
    :rtype: metric object
    """
    module_path = os.path.join(framework_path, "evaluation_metrics")
    module_name = defaults.metric_parameters[metric_name]["module"]
    module_path = os.path.join(module_path, module_name + ".py")
    Metric = importClass(metric_name, metric_name, module_path)
    metric = Metric(**metric_parameters)

    return metric


def choose_optimizer(optimizer):
    """
    Choose a surrogate model for Bayesian Optimization

    :param optimizer: list of setting of the BO experiment
    :type optimizer: Optimizer
    :return: surrogate model
    :rtype: scikit object
    """
    params_space_list = dimensions_aslist(optimizer.search_space)
    estimator = None
    # Choice of the surrogate model
    # Random forest
    if optimizer.surrogate_model == "RF":
        estimator = RandomForestRegressor(
            n_estimators=100, min_samples_leaf=3, random_state=optimizer.random_state)
    # Extra Tree
    elif optimizer.surrogate_model == "ET":
        estimator = ExtraTreesRegressor(
            n_estimators=100, min_samples_leaf=3, random_state=optimizer.random_state)
    # GP Minimize
    elif optimizer.surrogate_model == "GP":
        estimator = GaussianProcessRegressor(
            kernel=optimizer.kernel, random_state=optimizer.random_state)
        # Random Search
    elif optimizer.surrogate_model == "RS":
        estimator = "dummy"

    if estimator == "dummy":
        opt = skopt_optimizer(params_space_list, base_estimator=estimator,
                              acq_func=optimizer.acq_func,
                              acq_optimizer='sampling',
                              initial_point_generator=optimizer.initial_point_generator,
                              random_state=optimizer.random_state)
    else:
        opt = skopt_optimizer(params_space_list, base_estimator=estimator,
                              acq_func=optimizer.acq_func,
                              acq_optimizer='sampling',
                              n_initial_points=optimizer.n_random_starts,
                              initial_point_generator=optimizer.initial_point_generator,
                              # work only for version skopt 8.0!!!
                              acq_optimizer_kwargs={
                                  "n_points": 10000, "n_restarts_optimizer": 5, "n_jobs": 1},
                              acq_func_kwargs={"xi": 0.01, "kappa": 1.96},
                              random_state=optimizer.random_state)
    return opt


def convergence_res(values, optimization_type="minimize"):
    """
    Compute the list of values to plot the convergence plot (i.e. the best seen at each iteration)

    :param values: the result(s) for which to compute the convergence trace.
    :type values: list
    :param optimization_type: "minimize" if the problem is a minimization problem, "maximize" otherwise
    :type optimization_type: str
    :return: a list with the best min seen for each iteration
    :rtype: list
    """

    values2 = values.copy()

    if optimization_type == "minimize":
        for i in range(1, len(values2)):
            if values2[i] > values2[i - 1]:
                values2[i] = values2[i - 1]
    else:
        for i in range(1, len(values2)):
            if values2[i] < values2[i - 1]:
                values2[i] = values2[i - 1]
    return values2


def early_condition(values, n_stop, n_random):
    """
    Compute the early-stop criterium to stop or not the optimization.

    :param values: values obtained by Bayesian Optimization
    :type values: list
    :param n_stop: Range of points without improvement
    :type n_stop: int
    :param n_random: Random starting points
    :type n_random: int
    :return: 'True' if early stop condition reached, 'False' otherwise
    :rtype: bool
    """
    n_min_len = n_stop + n_random

    if len(values) >= n_min_len:
        values = convergence_res(values, optimization_type="minimize")
        worst = values[len(values) - n_stop]
        best = values[-1]
        diff = worst - best
        if diff == 0:
            return True

    return False


def plot_model_runs(model_runs, current_call, name_plot):
    """
    Save a boxplot of the data (Works only when optimization_runs is 1).

    :param model_runs: dictionary of all the model runs.
    :type model_runs: dict
    :param current_call: number of calls computed by BO
    :type current_call: int
    :param name_plot: Name of the plot
    :type name_plot: str
    """
    values = [model_runs["iteration_" + str(i)] for i in range(current_call + 1)]

    plt.ioff()
    plt.xlabel('number of calls')
    plt.grid(True)
    plt.boxplot(values)

    plt.savefig(name_plot + ".png")

    plt.close()


def plot_bayesian_optimization(values, name_plot,
                               log_scale=False, conv_max=True):
    """
    Save a convergence plot of the result of a Bayesian_optimization.

    :param values: List of objective function values
    :type values: list
    :param name_plot: Name of the plot
    :type name_plot: str
    :param log_scale: 'True' if you want a log scale for y-axis, 'False' otherwise
    :type log_scale: bool, optional
    :param conv_max: 'True' for a minimization problem, 'False' for a maximization problem
    :type conv_max: bool, optional
    """
    if conv_max:
        # minimization problem -->maximization problem
        values = [-val for val in values]
        media = convergence_res(values, optimization_type="maximize")
        xlabel = 'max f(x) after n calls'
    else:
        # minimization problem
        media = convergence_res(values, optimization_type="minimize")
        xlabel = 'min f(x) after n calls'

    array = [i for i in range(len(media))]

    plt.ioff()
    plt.plot(array, media, color='blue', label="res")

    if log_scale:
        plt.yscale('log')

    plt.ylabel(xlabel)
    plt.xlabel('Number of calls n')
    plt.legend(loc='best')
    plt.tight_layout()
    plt.grid(True)

    plt.savefig(name_plot + ".png")

    plt.close()


def convert_type(obj):
    """
    Convert a numpy object to a python object

    :param obj: object to be checked
    :type obj: numpy object
    :return: python object
    :rtype: python object
    """
    if isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    else:
        return obj


def check_instance(obj):
    """
    Check if a specific object con be inserted in the json file.

    :param obj: an object of the optimization to be saved
    :type obj: [str,float, int, bool, etc.]
    :return: 'True' if the object can be inserted in a json file, 'False' otherwise
    :rtype: bool
    """
    types = [str, float, int, bool]

    for t in types:
        if isinstance(obj, t):
            return True

    return False


def save_search_space(search_space):
    """
    Save the search space in the json file

    :param search_space: dictionary of the search space (scikit-optimize object)
    :type search_space: dict
    :return: dictionary for the seach space, which can be saved in a json file
    :rtype: dict
    """
    from skopt.space.space import Real, Categorical, Integer

    ss = dict()
    for key in list(search_space.keys()):
        if type(search_space[key]) == Real:
            ss[key] = ['Real', search_space[key].bounds, search_space[key].prior]
        elif type(search_space[key]) == Integer:
            ss[key] = ['Integer', search_space[key].bounds, search_space[key].prior]
        elif type(search_space[key]) == Categorical:
            ss[key] = ['Categorical', search_space[key].categories, search_space[key].prior]

    return ss


def load_search_space(search_space):
    """
    Load the search space from the json file

    :param search_space: dictionary of the search space (insertable in a json file)
    :type dict:
    :return: dictionary for the search space (for scikit optimize)
    :rtype: dict
    """
    from skopt.space.space import Real, Categorical, Integer

    ss = dict()
    for key in list(search_space.keys()):
        if search_space[key][0] == 'Real':
            ss[key] = Real(low=search_space[key][1][0], high=search_space[key][1][1], prior=search_space[key][2])
        elif search_space[key][0] == 'Integer':
            ss[key] = Integer(low=search_space[key][1][0], high=search_space[key][1][1], prior=search_space[key][2])
        elif search_space[key][0] == 'Categorical':
            ss[key] = Categorical(categories=search_space[key][1])

    return ss




from abc import ABC, abstractmethod


class AbstractMetric(ABC):
    """
    Class structure of a generic metric implementation
    """

    def __init__(self):
        """
        init metric
        """
        pass

    @abstractmethod
    def score(self, model_output):
        """
        Retrieves the score of the metric

        :param model_output: output of a topic model in the form of a dictionary. See model for details on
        the model output
        :type model_output: dict
        """
        pass

    def get_params(self):
        return [att for att in dir(self) if not att.startswith("_") and att != 'info' and att != 'score' and
                att != 'get_params']



from octis.evaluation_metrics.metrics import AbstractMetric
import octis.configuration.citations as citations
from sklearn.metrics import f1_score, precision_score, recall_score, accuracy_score
import numpy as np
from sklearn import svm
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from libsvm.svmutil import *


stored_average = None
stored_use_log = None
stored_scale = None
stored_kernel = None
stored_model_output_hash = None
stored_svm_results = [None, None]


class ClassificationScore(AbstractMetric):
    def __init__(self, dataset, average='micro', use_log=False, scale=True, kernel='linear', same_svm=False):
        AbstractMetric.__init__(self)

        self._train_document_representations = None
        self._test_document_representations = None

        self._labels = dataset.get_labels()
        self.average = average

        self.same_svm = same_svm

        self.use_log = use_log
        self.scale = scale

        self.kernel = kernel

    def score(self, model_output):
        self._train_document_representations = model_output["topic-document-matrix"].T
        self._test_document_representations = model_output["test-topic-document-matrix"].T

        if self.use_log:
            self._train_document_representations = np.log(
                self._train_document_representations)
            self._test_document_representations = np.log(
                self._test_document_representations)
        if self.scale:
            #scaler = MinMaxScaler()
            scaler2 = StandardScaler()
            x_train = scaler2.fit_transform(
                self._train_document_representations)
            x_test = scaler2.transform(self._test_document_representations)
        else:
            x_train = self._train_document_representations
            x_test = self._test_document_representations
        train_labels = [label for label in self._labels[:len(x_train)]]
        test_labels = [label for label in self._labels[-len(x_test):]]

        id2label, label2id = {}, {}
        for i, lab in enumerate(list(train_labels)):
            id2label[i] = lab
            label2id[lab] = i

        train_labels = [label2id[l] for l in train_labels]
        test_labels = [label2id[l] for l in test_labels]

        '''
        m = svm_train(train_labels, X_train, '-t 0')# -S 0 -K 2 -Z ')
        p_label, _, _ = svm_predict(test_labels, X_test, m)
        #print(len(X_test))
        #print(X_test.shape)
        return f1_score(test_labels, p_label, average=self.average)
        '''

        if self.kernel == 'linear':
            clf = svm.LinearSVC(verbose=False)
        else:
            clf = svm.SVC(kernel=self.kernel, verbose=False)
        clf.fit(x_train, train_labels)

        predicted_test_labels = clf.predict(x_test)

        return test_labels, predicted_test_labels


def compute_SVM_output(model_output, metric, super_metric):
    global stored_average
    global stored_use_log
    global stored_scale
    global stored_kernel
    global stored_svm_results
    global stored_model_output_hash

    model_output_hash = hash(str(model_output))

    test_labels = None
    predicted_test_labels = None
    flag = True

    if stored_average == metric.average and \
       stored_use_log == metric.use_log and \
       stored_scale == metric.scale and \
       stored_kernel == metric.kernel and \
       stored_model_output_hash == model_output_hash:
        test_labels, predicted_test_labels = stored_svm_results
    else:
        test_labels, predicted_test_labels = super_metric.score(model_output)

        stored_average = metric.average
        stored_use_log = metric.use_log
        stored_scale = metric.scale
        stored_kernel = metric.kernel
        stored_svm_results = [test_labels, predicted_test_labels]
        stored_model_output_hash = model_output_hash
        flag = False

    return [test_labels, predicted_test_labels, flag]


class F1Score(ClassificationScore):
    def __init__(self, dataset, average='micro', use_log=False, scale=True, kernel='linear', same_svm=False):
        super().__init__(dataset=dataset, average=average,
                         use_log=use_log, scale=scale, kernel=kernel, same_svm=same_svm)

    def info(self):
        return {
            "citation": citations.em_f1_score,
            "name": "F1 Score"
        }

    def score(self, model_output):
        """
        Retrieves the score of the metric

        Parameters
        ----------
        model_output : dictionary, output of the model. keys 'topic-document-matrix' and
                       'test-topic-document-matrix' are required.

        Returns
        -------
        score : score
        """
        test_labels, predicted_test_labels, self.same_svm = compute_SVM_output(model_output, self, super())
        return f1_score(test_labels, predicted_test_labels, average=self.average)


class PrecisionScore(ClassificationScore):

    def __init__(self, dataset, average='micro', use_log=False, scale=True, kernel='linear', same_svm=False):
        super().__init__(dataset=dataset, average=average,
                         use_log=use_log, scale=scale, kernel=kernel, same_svm=same_svm)

    def info(self):
        return {"citation": citations.em_f1_score, "name": "Precision"}

    def score(self, model_output):
        """
        Retrieves the score of the metric

        Parameters
        ----------
        model_output : dictionary, output of the model. 'topic-document-matrix' and
                       'test-topic-document-matrix' keys are required.

        Returns
        -------
        score : score
        """
        test_labels, predicted_test_labels, self.same_svm = compute_SVM_output(model_output, self, super())
        return precision_score(test_labels, predicted_test_labels, average=self.average)


class RecallScore(ClassificationScore):

    def __init__(self, dataset, average='micro', use_log=False, scale=True, kernel='linear', same_svm=False):
        super().__init__(dataset=dataset, average=average,
                         use_log=use_log, scale=scale, kernel=kernel, same_svm=same_svm)

    def info(self):
        return {"citation": citations.em_f1_score, "name": "Precision"}

    def score(self, model_output):
        """
        Retrieves the score of the metric

        Parameters
        ----------
        model_output : dictionary, output of the model. 'topic-document-matrix' and
                       'test-topic-document-matrix' keys are required.

        Returns
        -------
        score : score
        """
        test_labels, predicted_test_labels, self.same_svm = compute_SVM_output(model_output, self, super())
        return recall_score(test_labels, predicted_test_labels, average=self.average)

class AccuracyScore(ClassificationScore):

    def __init__(self, dataset, average='micro', use_log=False, scale=True, kernel='linear', same_svm=False):
        super().__init__(dataset=dataset, average=average,
                         use_log=use_log, scale=scale, kernel=kernel, same_svm=same_svm)

    def info(self):
        return {"citation": citations.em_f1_score, "name": "Precision"}

    def score(self, model_output):
        """
        Retrieves the score of the metric

        Parameters
        ----------
        model_output : dictionary, output of the model. 'topic-document-matrix' and
                       'test-topic-document-matrix' keys are required.

        Returns
        -------
        score : score
        """
        test_labels, predicted_test_labels, self.same_svm = compute_SVM_output(model_output, self, super())
        return accuracy_score(test_labels, predicted_test_labels)


#!/usr/bin/env python

"""Tests for `octis` package."""

import pytest

from click.testing import CliRunner
from octis.evaluation_metrics.classification_metrics import F1Score

from octis.evaluation_metrics.coherence_metrics import *
from octis.dataset.dataset import Dataset

import os
from octis.preprocessing.preprocessing import Preprocessing

from octis.dataset.downloader import get_data_home, _pkl_filepath

@pytest.fixture
def root_dir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def data_dir(root_dir):
    return root_dir + "/../preprocessed_datasets/"


def test_preprocessing_custom_stops(data_dir):
    texts_path = data_dir+"/sample_texts/unprepr_docs.txt"
    p = Preprocessing(vocabulary=None, max_features=None, remove_punctuation=True, punctuation=".,?:",
                      lemmatize=False,  stopword_list=['am', 'are', 'this', 'that'],
                      min_chars=2, min_words_docs=5,min_df=0.0001)
    dataset = p.preprocess_dataset(
        documents_path=texts_path,
    )

    dataset.save(data_dir+"/sample_texts/")
    dataset.load_custom_dataset_from_folder(data_dir + "/sample_texts")

def test_preprocessing_english_stops(data_dir):
    texts_path = data_dir+"/sample_texts/unprepr_docs.txt"
    p = Preprocessing(vocabulary=None, max_features=None, remove_punctuation=True,
                      lemmatize=False,  stopword_list='english',
                      min_chars=2, min_words_docs=1)
    dataset = p.preprocess_dataset(
        documents_path=texts_path,
    )

    dataset.save(data_dir+"/sample_texts/")
    dataset.load_custom_dataset_from_folder(data_dir + "/sample_texts")


def test_load_20ng():
    data_home = get_data_home(data_home=None)
    cache_path = _pkl_filepath(data_home, "20NewsGroup" + ".pkz")
    if os.path.exists(cache_path):
        os.remove(cache_path)

    dataset = Dataset()
    dataset.fetch_dataset("20NewsGroup")
    assert len(dataset.get_corpus()) == 16309
    assert len(dataset.get_labels()) == 16309
    assert os.path.exists(cache_path)

    dataset = Dataset()
    dataset.fetch_dataset("20NewsGroup")
    assert len(dataset.get_corpus()) == 16309


def test_load_M10():
    dataset = Dataset()
    dataset.fetch_dataset("M10")
    assert len(set(dataset.get_labels())) == 10


def test_partitions_fetch():
    dataset = Dataset()
    dataset.fetch_dataset("M10")
    partitions = dataset.get_partitioned_corpus()
    assert len(partitions[0]) == 5847
    assert len(partitions[1]) == 1254


def test_partitions_custom(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir+"M10")
    partitions = dataset.get_partitioned_corpus()
    assert len(partitions[0]) == 5847
    assert len(partitions[1]) == 1254



from pathlib import Path




'''
# MODELS PARAMETERS #
'''
model_descriptions = {'LDA': {'name': 'Latent Dirichlet Allocation (LDA)',
                              'citation': 'David M. Blei, Andrew Y. Ng, Michael I. Jordan: Latent Dirichlet Allocation. '
                                          'NIPS 2001: 601-608'},
                      'ETM': {'name': 'Embedded Topic Models (ETM)',
                              'citation': 'Adji Bousso Dieng, Francisco J. R. Ruiz, David M. Blei: Topic Modeling in '
                                          'Embedding Spaces. Trans. Assoc. Comput. Linguistics 8: 439-453 (2020)'},
                      'LSI': {'name': 'Latent Semantic Indexing',
                              'citation': 'Deerwester, S., et al, Improving Information Retrieval with Latent Semantic '
                                          'Indexing, Proceedings of the 51st Annual Meeting of the American Society for '
                                          'Information Science 25, 1988, pp. 36–40'},
                      'ProdLDA': {'name': 'Product-of-Experts LDA',
                                  'citation': 'Akash Srivastava, Charles Sutton: Autoencoding Variational Inference For '
                                              'Topic Models. ICLR (Poster) 2017'},
                      'NeuralLDA': {'name': 'Neural LDA',
                                    'citation': 'Akash Srivastava, Charles Sutton: Autoencoding Variational Inference For '
                                                'Topic Models. ICLR (Poster) 2017'},
                      'NMF': {'name': 'Non-negative Matrix Factorization',
                              'citation': 'Daniel D. Lee & H. Sebastian Seung (2001). Algorithms for Non-negative Matrix '
                                          'Factorization. Advances in Neural Information Processing Systems 13: '
                                          'Proceedings of the 2000 Conference. MIT Press. pp. 556–562.'}}

model_hyperparameters = {
    'LDA': {
        'alpha': {'type': 'Real', 'default_value': 0.1, 'min_value': 1e-4, 'max_value': 20,
                  'step': 1e-4, 'alternative_name': '&alpha;',
                  'description': 'symmetric Dirichlet prior controlling the document-topic distribution. '
                                 'Low values (<1) place more weight on having each document composed of '
                                 'only a few dominant topics (type: real)'},
        'eta': {'type': 'Real', 'default_value': 0.1, 'min_value': 1e-4, 'max_value': 20, 'step': 1e-4,
                'alternative_name': '&beta;', 'description': 'symmetric Dirichlet prior over the vocabulary. '
                                                             'Low values (<1) place more weight on having each topic'
                                                             ' composed of only a few dominant words (type: real)'},
        'num_topics': {'type': 'Integer', 'default_value': 10, 'min_value': 2,
                       'max_value': 200, 'step': 1, 'alternative_name': 'number of topics', 'description':
                           'number of topics to discover (integer)'},
        'passes': {'type': 'Integer', 'default_value': 1, 'min_value': 1, 'max_value': 10, 'step': 1,
                   'description': 'Number of passes through the corpus during training', 'alternative_name':
                       'Number of passes'},
        'iterations': {'type': 'Integer', 'default_value': 50, 'min_value': 5,
                       'max_value': 2000, 'step': 1,
                       'description': 'Number of iterations of the topic model algorithm (integer)'}},
    'ETM': {'num_topics': {'type': 'Integer', 'default_value': 10, 'min_value': 2,
                           'max_value': 200, 'step': 1, 'alternative_name': 'number of topics', 'description':
                               'number of topics to discover (integer)'},
            'num_epochs': {'type': 'Integer', 'default_value': 50, 'min_value': 5,
                           'max_value': 300, 'step': 1, 'alternative_name': 'number of training epochs', 'description':
                               'number of training epochs (type: integer)'},
            't_hidden_size':
                {'type': 'Categorical', 'default_value': 800,
                 'possible_values': [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000],
                 'alternative_name': 'hidden size',
                 'description': 'size of the deepest hidden layer (type: categorical)'},
            'rho_size':
                {'type': 'Categorical', 'default_value': 300, 'possible_values': [100, 200, 300, 400, 500],
                 'alternative_name': 'rho size',
                 'description': 'size of the word embeddings to learn (type: categorical)'},

            # 'embedding_size': {'type': 'Categorical', 'default_value': 300,
            #                  'alternative_name': 'pretrained embeddings size',
            #                  'description': 'size of the word embeddings that have been uploaded (type: categorical)'},

            'activation': {'type': 'Categorical', 'default_value': 'relu',
                           'possible_values': ['relu', 'tanh', 'sigmoid', 'softplus', 'rrelu', 'leakyrelu', 'elu',
                                               'selu', 'glu'],
                           'alternative_name': 'activation function',
                           'description': 'activation function applied to the hidden layer (type: categorical)'},
            'dropout': {'type': 'Real', 'default_value': 0.5, 'min_value': 0.0, 'max_value': 0.95, 'step': 0.05,
                        'description': 'Dropout applied to the hidden layer',
                        'alternative_name': 'dropout'},
            'lr': {'type': 'Real', 'default_value': 0.005, 'min_value': 1e-6, 'max_value': 0.1, 'step': 1e-6,
                   'description': 'Learning rate for the training of the network',
                   'alternative_name': 'learning rate'},
            'batch_size': {'type': 'Categorical', 'default_value': 200,
                           'possible_values': [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000],
                           'alternative_name': 'batch size',
                           'description': 'batch size (type: categorical)'},
            'bow_norm': {'type': 'Categorical', 'default_value': 1, 'possible_values': [0, 1],
                         'alternative_name': 'bow normalization', 'description':
                             'Whether to normalize (1) or not normalize (0) the BOW (bag-of-words) documents.'},
            'wdecay': {'type': 'Real', 'default_value': 1.2e-6, 'min_value': 0, 'max_value': 1, 'step': 1e-6,
                       'alternative_name': 'weight decay', 'description':
                           'Weight decay factor.'},

            'optimizer': {'type': 'Categorical', 'default_value': 'sgd',
                          'possible_values': ['adam', 'adagrad', 'adadelta', 'rmsprop', 'asgd', 'sgd'],
                          'alternative_name': 'optimizer',
                          'description': 'optimizer algorithm (type: categorical)'},

            },
    'LSI': {
        'decay': {'type': 'Real', 'default_value': 1.0, 'min_value': 0.0, 'max_value': 1.0, 'step': 0.1,
                  'description': 'Weight of existing observations relatively to new ones',
                  'alternative_name': 'decay'},
        'onepass': {'type': 'Categorical', 'default_value': True, 'possible_values': [True, False],
                    'description': 'Whether the one-pass algorithm should be used for training. '
                                   'Pass False to force a multi-pass stochastic algorithm.',
                    'alternative_name': 'one-pass algorithm'},
        'num_topics': {'type': 'Integer', 'default_value': 10, 'min_value': 2,
                       'max_value': 200, 'step': 1, 'alternative_name': 'number of topics', 'description':
                           'number of topics to discover'},
        'power_iters': {'type': 'Integer', 'default_value': 1, 'min_value': 1, 'max_value': 5, 'step': 1,
                        'description': 'Number of power iteration steps to be used. Increasing the number of power '
                                       'iterations improves accuracy, but lowers performance', 'alternative_name':
                        'Number of power iteration steps'},
        'extra_samples': {'type': 'Integer', 'default_value': 100, 'min_value': 0, 'max_value': 500, 'step': 1,
                          'description': 'Extra samples to be used besides the rank k. Can improve accuracy',
                          'alternative_name': 'Number of extra samples'}},
    'NMF': {
        'num_topics': {'type': 'Integer', 'default_value': 10, 'min_value': 2, 'max_value': 200, 'step': 1,
                       'alternative_name': 'number of topics', 'description': 'number of topics to discover'
                       },
        'passes': {'type': 'Integer', 'default_value': 1, 'min_value': 1, 'max_value': 10, 'step': 1,
                   'description': 'Number of passes through the corpus during training', 'alternative_name':
                       'Number of passes'},
        'normalize': {'type': 'Categorical', 'default_value': True, 'possible_values': [True, False],
                      'description': 'Whether to normalize the result'},
        'eval_every': {'type': 'Integer', 'default_value': 10, 'min_value': 1, 'max_value': 50, 'step': 1,
                       'description': 'Number of batches after which l2 norm of (v - Wh) is computed. '
                                      'Decreases performance if set too low.', 'alternative_name': 'evaluate every'},
        'kappa': {'type': 'Real', 'default_value': 1.0, 'min_value': 0.1, 'max_value': 5.0, 'step': 0.1,
                  'alternative_name': 'k',
                  'description': 'Gradient descent step size. Larger value makes the model train faster, but could lead to non-convergence if set too large.'},
        'w_max_iter': {'type': 'Integer', 'default_value': 200, 'min_value': 5, 'max_value': 1000, 'step': 1,
                       'alternative_name': 'Maximum iterations for W', 'description':
                           'Maximum number of iterations to train the matrix W per each batch (type: integer'},
        'h_max_iter': {'type': 'Integer', 'default_value': 50, 'min_value': 5, 'max_value': 1000, 'step': 1,
                       'alternative_name': 'Maximum iterations for H', 'description':
                           'Maximum number of iterations to train the matrix H per each batch (type: integer)'
                       },
        'w_stop_condition': {'type': 'Real', 'default_value': 0.0001, 'min_value': 1E-6, 'max_value': 0.1,
                             'step': 1E-6, 'alternative_name': 'stopping condition for W',
                             'description': 'If error difference gets less than that, training of W stops for '
                                            'the current batch. (type: real)'},
        'h_stop_condition': {'type': 'Real', 'default_value': 0.0001, 'min_value': 1E-6, 'max_value': 0.1,
                             'step': 1E-6, 'alternative_name': 'stopping condition for H',
                             'description': 'If error difference gets less than that, training of H stops for '
                                            'the current batch. (type: real)'}},
    'NMF_scikit': {
        'num_topics': {'type': 'Integer', 'default_value': 10, 'min_value': 2, 'max_value': 200, 'step': 1,
                       'alternative_name': 'number of topics', 'description': 'number of topics to discover'
                       },
        'init': {'type': 'Categorical', 'default_value': None,
                 'possible_values': [None, 'random', 'nndsvd', 'nndsvda', 'nndsvdar']},
        'alpha': {'type': 'Real', 'default_value': 0.0, 'min_value': 0.0, 'max_value': 1.0, 'step': 0.1},
        'l1_ratio': {'type': 'Real', 'default_value': 0.0, 'min_value': 0, 'max_value': 1, 'step': 0.1}},
    'NeuralLDA': {
        'num_topics': {'type': 'Integer', 'default_value': 10, 'min_value': 2, 'max_value': 200, 'step': 1,
                       'alternative_name': 'number of topics', 'description': 'number of topics to discover'
                       },
        'activation': {'type': 'Categorical', 'default_value': 'softplus',
                       'possible_values': ['softplus', 'relu', 'sigmoid', 'swish', 'tanh', 'leakyrelu',
                                           'rrelu', 'elu', 'selu'],
                       'alternative_name': 'activation function',
                       'description': 'activation function applied to the hidden layer (type: categorical)'},
        'dropout': {'type': 'Real', 'default_value': 0.2, 'min_value': 0.0, 'max_value': 0.95, 'step': 0.05,
                    'description': 'Dropout applied to the hidden layer',
                    'alternative_name': 'dropout'},
        'batch_size': {'type': 'Categorical', 'default_value': 200,
                       'possible_values': [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000],
                       'alternative_name': 'batch size',
                       'description': 'batch size (type: categorical)'},
        'lr': {'type': 'Real', 'default_value': 2e-3, 'min_value': 1e-6, 'max_value': 0.1, 'step': 1e-6,
               'description': 'Learning rate for the training of the network',
               'alternative_name': 'learning rate'},
        'momentum': {'type': 'Real', 'default_value': 0.99, 'min_value': 0.01, 'max_value': 1, 'step': 0.01,
                     'description': 'Momentum',
                     'alternative_name': 'Momentum'},
        'solver': {'type': 'Categorical', 'default_value': 'sgd',
                   'possible_values': ['adagrad', 'adam', 'sgd', 'adadelta', 'rmsprop'],
                   'alternative_name': 'optimizer',
                   'description': 'optimizer algorithm (type: categorical)'},
        'num_epochs': {'type': 'Integer', 'default_value': 50, 'min_value': 5,
                       'max_value': 300, 'step': 1, 'alternative_name': 'number of training epochs', 'description':
                           'number of training epochs (type: integer)'},
        'num_neurons':
            {'type': 'Categorical', 'default_value': 800,
             'possible_values': [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000],
             'alternative_name': 'Number of neurons',
             'description': 'Number of neurons for each layer of the network (type: integer)'},

        'num_layers':
            {'type': 'Categorical', 'default_value': 2,
             'possible_values': [1, 2, 3, 4, 5],
             'alternative_name': 'Number of layers',
             'description': 'Number of layers of the network (type: integer)'}},
    'ProdLDA': {
        'num_topics': {'type': 'Integer', 'default_value': 10, 'min_value': 2, 'max_value': 200, 'step': 1,
                       'alternative_name': 'number of topics', 'description': 'number of topics to discover'
                       },
        'activation': {'type': 'Categorical', 'default_value': 'softplus',
                       'possible_values': ['softplus', 'relu', 'sigmoid', 'swish', 'tanh', 'leakyrelu',
                                           'rrelu', 'elu', 'selu'],
                       'alternative_name': 'activation function',
                       'description': 'activation function applied to the hidden layer (type: categorical)'},
        'dropout': {'type': 'Real', 'default_value': 0.2, 'min_value': 0.0, 'max_value': 0.95, 'step': 0.05,
                    'description': 'Dropout applied to the hidden layer',
                    'alternative_name': 'dropout'},
        'batch_size': {'type': 'Categorical', 'default_value': 200,
                       'possible_values': [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000],
                       'alternative_name': 'batch size',
                       'description': 'batch size (type: categorical)'},
        'lr': {'type': 'Real', 'default_value': 2e-3, 'min_value': 1e-6, 'max_value': 0.1, 'step': 1e-6,
               'description': 'Learning rate for the training of the network',
               'alternative_name': 'learning rate'},
        'momentum': {'type': 'Real', 'default_value': 0.99, 'min_value': 0.01, 'max_value': 1, 'step': 0.01,
                     'description': 'Momentum',
                     'alternative_name': 'Momentum'},
        'solver': {'type': 'Categorical', 'default_value': 'sgd',
                   'possible_values': ['adagrad', 'adam', 'sgd', 'adadelta', 'rmsprop'],
                   'alternative_name': 'optimizer',
                   'description': 'optimizer algorithm (type: categorical)'},
        'num_epochs': {'type': 'Integer', 'default_value': 50, 'min_value': 5,
                       'max_value': 300, 'step': 1, 'alternative_name': 'number of training epochs', 'description':
                           'number of training epochs (type: integer)'},
        'num_neurons':
            {'type': 'Categorical', 'default_value': 800,
             'possible_values': [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000],
             'alternative_name': 'Number of neurons',
             'description': 'Number of neurons for each layer of the network (type: integer)'},

        'num_layers':
            {'type': 'Categorical', 'default_value': 2,
             'possible_values': [1, 2, 3, 4, 5],
             'alternative_name': 'Number of layers',
             'description': 'Number of layers of the network (type: integer)'}
    }
}

'''
# METRIC PARAMETERS #
'''
metric_parameters = {
    # coherence
    "Coherence": {
        "name": "Coherence", "module": "coherence_metrics",
        "texts": {"type": "String", "default_value": "use dataset texts"},
        "topk": {"type": "Integer", "default_value": 10, "min_value": 5, "max_value": 30},
        "measure": {"type": "Categorical", "default_value": "c_npmi",
                    "possible_values": ['u_mass', 'c_v', 'c_uci', 'c_npmi']}},
    # diversity
    "TopicDiversity": {
        "name": "% Unique words",
        "module": "diversity_metrics",
        "topk": {"type": "Integer", "default_value": 10, "min_value": 5,
                 "max_value": 30}},

    "InvertedRBO": {"name": "IRBO", "module": "diversity_metrics",
                    "topk": {"type": "Integer", "default_value": 10, "min_value": 5,
                             "max_value": 30},
                    "weight": {"type": "Real", "default_value": 0.9, "min_value": 0.0,
                               "max_value": 1.0}},
    # divergences
    "KL_uniform": {"name": "KL-U", "module": "topic_significance_metrics",
                   "description": "KL-Uniform: It measures the distance (measured using the KL-divergence) "
                                  "of each topic-word distribution "
                                  "from the uniform distribution over the words. Significant topics are "
                                  "supposed to be skewed towards a few coherent and related words and "
                                  "distant from the uniform distribution."},
    "KL_background": {"name": "KL-B", "module": "topic_significance_metrics",
                      "description": "It measures the distance of a topic k to a “background” topic, which is a "
                                     "generic topic that is found equally probable in all the documents. Meaningful"
                                     " topics appear in a small subset of the data,thus higher values of KL–B are"
                                     " preferred"},
    "KL_vacuous": {"name": "KL-V", "module": "topic_significance_metrics",
                   "description": "It measures the distance between each topic-word distribution and the "
                                  "empirical word distribution of the whole dataset, also called “vacuous” "
                                  "distribution. The closer the word-topic distribution is to the empirical "
                                  "distribution of the sample, the less its significance is expected to be"},

    # classification
    "F1Score": {"name": "F1Score", "module": "classification_metrics",
                "dataset": {"type": "String", "default_value": "use selected dataset"},
                'average': {"type": "Categorical", "default_value": "micro",
                            "possible_values": ['binary', 'micro', None, 'macro', 'samples', 'weighted']
                            }}

}

'''
# OPTIMIZATION PARAMETERS #
'''

optimization_parameters = {
    "surrogate_models": [{"name": "Gaussian process", "id": "GP"},
                         {"name": "Random forest", "id": "RF"}],
    "acquisition_functions": [{"name": "Upper confidence bound", "id": "LCB"},
                              {"name": "Expected improvement", "id": "EI"}]
}

'''
# PARAMETERS INFO #
'''
HDP_hyperparameters_info = """
max_chunks (int, optional) – Upper bound on how many chunks to process. It wraps around corpus beginning in another corpus pass, if there are not enough chunks in the corpus. \n

max_time (int, optional) – Upper bound on time (in seconds) for which model will be trained. \n

chunksize (int, optional) – Number of documents in one chuck. \n

kappa (float,optional) – Learning parameter which acts as exponential decay factor to influence extent of learning from each batch. \n

tau (float, optional) – Learning parameter which down-weights early iterations of documents. \n

K (int, optional) – Second level truncation level \n

T (int, optional) – Top level truncation level \n

alpha (int, optional) – Second level concentration \n

gamma (int, optional) – First level concentration \n

eta (float, optional) – The topic Dirichlet \n

scale (float, optional) – Weights information from the mini-chunk of corpus to calculate rhot. \n

var_converge (float, optional) – Lower bound on the right side of convergence. Used when updating variational parameters for a single document.
"""

LDA_hyperparameters_info = """
num_topics (int, optional) – The number of requested latent topics to be extracted from the training corpus. \n

distributed (bool, optional) – Whether distributed computing should be used to accelerate training. \n

chunksize (int, optional) – Number of documents to be used in each training chunk. \n

passes (int, optional) – Number of passes through the corpus during training. \n

update_every (int, optional) – Number of documents to be iterated through for each update. Set to 0 for batch learning, > 1 for online iterative learning. \n

alpha ({numpy.ndarray, str}, optional) – Can be set to an 1D array of length equal to the number of expected topics that expresses our a-priori belief for the each topics’ probability. Alternatively default prior selecting strategies can be employed by supplying a string:

’asymmetric’: Uses a fixed normalized asymmetric prior of 1.0 / topicno.

’auto’: Learns an asymmetric prior from the corpus (not available if distributed==True). \n

eta ({float, np.array, str}, optional) – A-priori belief on word probability, this can be:

scalar for a symmetric prior over topic/word probability,

vector of length num_words to denote an asymmetric user defined probability for each word,

matrix of shape (num_topics, num_words) to assign a probability for each word-topic combination,

the string ‘auto’ to learn the asymmetric prior from the data. \n

decay (float, optional) – A number between (0.5, 1] to weight what percentage of the previous lambda value is forgotten when each new document is examined. Corresponds to Kappa from Matthew D. Hoffman, David M. Blei, Francis Bach: “Online Learning for Latent Dirichlet Allocation NIPS’10”. \n

offset (float, optional) –

Hyper-parameter that controls how much we will slow down the first steps the first few iterations.
Corresponds to Tau_0 from Matthew D. Hoffman, David M. Blei, Francis Bach: “Online Learning for Latent Dirichlet Allocation NIPS’10”. \n

eval_every (int, optional) – Log perplexity is estimated every that many updates. Setting this to one slows
down training by ~2x. \n

iterations (int, optional) – Maximum number of iterations through the corpus when inferring the topic
distribution of a corpus. \n

gamma_threshold (float, optional) – Minimum change in the value of the gamma parameters to continue iterating. \n

minimum_probability (float, optional) – Topics with a probability lower than this threshold will be filtered out. \n

random_state ({np.random.RandomState, int}, optional) – Either a randomState object or a seed to generate one. Useful
for reproducibility. \n

minimum_phi_value (float, optional) – if per_word_topics is True, this represents a lower bound on the term
probabilities. \n

per_word_topics (bool) – If True, the model also computes a list of topics, sorted in descending order of most likely
topics for each word, along with their phi values multiplied by the feature length (i.e. word count). \n
"""

LSI_hyperparameters_info = """
num_topics (int, optional) – Number of requested factors (latent dimensions)\n

chunksize (int, optional) – Number of documents to be used in each training chunk. \n

decay (float, optional) – Weight of existing observations relatively to new ones. \n

distributed (bool, optional) – If True - distributed mode (parallel execution on several machines) will be used. \n

onepass (bool, optional) – Whether the one-pass algorithm should be used for training. Pass False to force a multi-pass
 stochastic algorithm. \n

power_iters (int, optional) – Number of power iteration steps to be used. Increasing the number of power iterations
improves accuracy, but lowers performance \n

extra_samples (int, optional) – Extra samples to be used besides the rank k. Can improve accuracy.
"""

NMF_gensim_hyperparameters_info = """
num_topics (int, optional) – Number of topics to extract. \n

chunksize (int, optional) – Number of documents to be used in each training chunk. \n

passes (int, optional) – Number of full passes over the training corpus. Leave at default passes=1 if your input is an
 iterator. \n

kappa (float, optional) – Gradient descent step size. Larger value makes the model train faster, but could lead to
non-convergence if set too large. \n

minimum_probability – If normalize is True, topics with smaller probabilities are filtered out. If normalize is False,
 topics with smaller factors are filtered out. If set to None, a value of 1e-8 is used to prevent 0s. \n

w_max_iter (int, optional) – Maximum number of iterations to train W per each batch. \n

w_stop_condition (float, optional) – If error difference gets less than that, training of W stops for the current
batch. \n

h_max_iter (int, optional) – Maximum number of iterations to train h per each batch. \n

h_stop_condition (float) – If error difference gets less than that, training of h stops for the current batch. \n

eval_every (int, optional) – Number of batches after which l2 norm of (v - Wh) is computed. Decreases performance if
set too low. \n

normalize (bool or None, optional) – Whether to normalize the result. \n

random_state ({np.random.RandomState, int}, optional) – Seed for random generator. Needed for reproducibility. \n
"""

NMF_scikit_hyperparameters_info = """
num_topics (int) – Number of topics to extract. \n

init (string, optional) – Method used to initialize the procedure. Default: None. Valid options:

    None: ‘nndsvd’ if n_components <= min(n_samples, n_features),
otherwise random.

    ‘random’: non-negative random matrices, scaled with:
sqrt(X.mean() / n_components)

    ‘nndsvd’: Nonnegative Double Singular Value Decomposition (NNDSVD)
initialization (better for sparseness)

    ‘nndsvda’: NNDSVD with zeros filled with the average of X
(better when sparsity is not desired)

    ‘nndsvdar’: NNDSVD with zeros filled with small random values
(generally faster, less accurate alternative to NNDSVDa for when sparsity is not desired) \n

alpha (double, optional) – Constant that multiplies the regularization terms. Set it to zero to have no regularization. \n

l1_ratio (double, optional) – The regularization mixing parameter, with 0 <= l1_ratio <= 1. For l1_ratio = 0 the penalty is an elementwise L2 penalty (aka Frobenius Norm). For l1_ratio = 1 it is an elementwise L1 penalty. For 0 < l1_ratio < 1, the penalty is a combination of L1 and L2.
"""



#!/usr/bin/env python

"""Tests for `octis` package."""

import pytest

from click.testing import CliRunner

from octis.dataset.dataset import Dataset
from octis.models.LDA import LDA
from octis.models.LDA_tomopy import LDA_tomopy as LDATOMOTO
from octis.models.ETM import ETM
from octis.models.CTM import CTM
from octis.models.NMF import NMF
from octis.models.NMF_scikit import NMF_scikit
from octis.models.ProdLDA import ProdLDA

import numpy as np
import os


@pytest.fixture
def root_dir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def data_dir(root_dir):
    return root_dir + "/../preprocessed_datasets/"


def test_model_output_lda(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = LDA(num_topics=num_topics, iterations=5)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_etm(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ETM(num_topics=num_topics, num_epochs=5)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_etm_not_partitioned(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ETM(num_topics=num_topics, num_epochs=5, use_partitions=False)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' not in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(dataset.get_corpus()))


def test_model_output_lda_tomotopy(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = LDATOMOTO(num_topics=num_topics, alpha=0.1)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_nmf(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = NMF(num_topics=num_topics, w_max_iter=10, h_max_iter=10, use_partitions=True)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_nmf_scikit(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = NMF_scikit(num_topics=num_topics, use_partitions=True)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_ctm_zeroshot(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = CTM(num_topics=num_topics, num_epochs=5, inference_type='zeroshot')
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_ctm_combined(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = CTM(num_topics=num_topics, num_epochs=5, inference_type='combined')
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_ctm_combined_not_partition(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = CTM(num_topics=num_topics, num_epochs=5, inference_type='combined',use_partitions=False,
                bert_path='./not_part')
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' not in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(dataset.get_corpus()))


def test_model_output_prodlda(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ProdLDA(num_topics=num_topics, num_epochs=5)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_prodlda_not_partitioned(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ProdLDA(num_topics=num_topics, num_epochs=5, use_partitions=False)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' not in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(dataset.get_corpus()))

